/**
 * @author Martin Heusel
 * 
 * Copyright (C) 2008 Martin Heusel (mhe@bioinf.jku.at),
 *                      
 * Johannes Kepler University, Linz, Austria
 * Institute of Bioinformatics.
 * The software is maintained by Martin Heusel.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * If you use this library, please cite:
 *
 * @article{SeppHochreiter07152007,
 *   author = {Hochreiter, Sepp and Heusel, Martin and Obermayer, Klaus},
 *   title = {{Fast Model-based Protein Homology Detection without Alignment}},
 *   journal = {Bioinformatics},
 *   volume = {23},
 *   number = {14},
 *   pages = {1728-1736},
 *   doi = {doi:10.1093/bioinformatics/btm247},
 *   year = {2007},
 *   URL = {http://bioinformatics.oxfordjournals.org/cgi/content/abstract/btm247v1},
 *   eprint = {http://bioinformatics.oxfordjournals.org/cgi/reprint/btm247v1}
 * }
 * 
 * $Id: JLSTM.java 237 2009-03-03 10:27:03Z mhe $
 *
 */

package at.jku.bioinf.jlstm;

import java.util.ArrayList;
import java.util.Properties;

/**
 * Main class
 * 
 * @author mhe
 *
 */
public class JLSTM {
	
	// Property file
	static Properties properties;
	
	// Data containers for the sequences
	static ArrayList<LSTMSequence> seqsPosTrain;
	static ArrayList<LSTMSequence> seqsNegTrain;
	static ArrayList<LSTMSequence> seqsPosTest;
	static ArrayList<LSTMSequence> seqsNegTest;
	
	static ArrayList<LSTMSequence> seqsTrain;
	static ArrayList<LSTMSequence> seqsTest;
	
	static int windowSize;
	
	/* Number of symbols. 20 or 23 for Proteins and 4 for DNA. */
    static int nSymbols;
    
    /* Index of output unit */
    static int out_mod;
    
    /* Number of LSTM blocks */
    static int num_blocks;
    
    /* Index of last input unit */
    static int in_mod;
    
    /* Bias unit */
    static int in_mod_b;
    
    /* Number of inputs for local coding */
    static int in_nn_mod;
    
    /* Index of last LSTM unit */
    static int cell_mod;
    
    /* Number of units */
    static int ges_mod;

    /* Blocksize for every block */
    static int[] blockSize;
    
    /* Range for random init of weights */
    static double init_range;
    
    /* Target values */
    static double targetpositive;
    static double targetnegative;
    
    /* Biases and initial weights */
    static double[] bias_inp;
    static double[] bias_out;
    static double[] bias_memin;
    static double[] weight_memout;
    static double outputbias;
    
    /* Learning rate */
    static double alpha;
    
    /* maximum number of epochs */
    static int maxepochs;
    
    /* After testnepochs a test is performed */
    static int testnepochs;
    
    /* After writeweightsafternepochs the weight matrix is written */
    static int writeweightsafternepochs;
    
    /* Write an additional better readable weight matrix? For debugging
     * and investigation.
     */
    static int humanreadableweightmatrix;
    
    /* Additional ROCN (e.g. N = 50) to be computed */
    static int rocn;

    /* Weight matrix */
    static WeightMatrix wm;
    
    static String weightfile;
    
    static String propertiesfile;
    
    /* test flag */
    static boolean test;
    
    static boolean loadweights;
    
    /* Input matrix */
    static InputMatrix im;
    
    /* LSTM core */
    static LSTM lstm;
	
    /* Number of threads */
    static int numberThreads;
	
	/**
	 * Read and set properties from file
	 * 
	 */
	static void readProperties() {
	    
	    System.err.println();
        System.err.print("reading properties.. ");
		
	    // Read properties
	    properties = JLSTMProperties.readProperties(propertiesfile);

        windowSize    = Integer.parseInt(properties.getProperty("windowsize"));
        nSymbols      = Integer.parseInt(properties.getProperty("numberofsymbols"));
        out_mod       = Integer.parseInt(properties.getProperty("numberoutputunits"));
        num_blocks    = Integer.parseInt(properties.getProperty("numbermemorycellblocks"));
        init_range    = Double.parseDouble(properties.getProperty("halfintervallengthforinit"));
        
        targetpositive  = Double.parseDouble(properties.getProperty("targetvaluepos"));
        targetnegative  = Double.parseDouble(properties.getProperty("targetvalueneg"));
        
        alpha         = Double.parseDouble(properties.getProperty("learningrate"));
        
        maxepochs     = Integer.parseInt(properties.getProperty("maxnumberofepochs"));
        
        testnepochs   = Integer.parseInt(properties.getProperty("performingtestafterNpochs"));
        
        writeweightsafternepochs  = Integer.parseInt(properties.getProperty("writeweightafterNepochs"));
        humanreadableweightmatrix = Integer.parseInt(properties.getProperty("humanreadableweightmatrix"));
                
        rocn          = Integer.parseInt(properties.getProperty("rocn"));
        
        // Blocksizes
        blockSize = new int[num_blocks];
        
        for (int i = 0; i < num_blocks; i++) {
            blockSize[i] = Integer.parseInt(properties.getProperty("blocksize" + String.valueOf(i + 1)));
        }
        
        // Input bias
        bias_memin = new double[num_blocks];
        
        for (int i = 0; i < num_blocks; i++) {
            bias_memin[i] = Double.parseDouble(properties.getProperty("initialinputbias" + String.valueOf(i + 1)));
        }

        // Inputgate biases
        bias_inp = new double[num_blocks];

        for (int i = 0; i < num_blocks; i++) {
            bias_inp[i] = Double.parseDouble(properties.getProperty("initialinputgatebias" + String.valueOf(i + 1)));
        }

        // Outputgate biases
        bias_out = new double[num_blocks];
        
        for (int i = 0; i < num_blocks; i++) {
            bias_out[i] = Double.parseDouble(properties.getProperty("initialoutputgatebias" + String.valueOf(i + 1)));
        }
        
        // Output weights
        weight_memout = new double[num_blocks];
        
        for (int i = 0; i < num_blocks; i++) {
            weight_memout[i] = Double.parseDouble(properties.getProperty("initialoutputweight" + String.valueOf(i + 1)));
        }

        // Output bias
        outputbias = Double.parseDouble(properties.getProperty("outputbias"));
        
        System.err.println("okay");
		
	}
	
	/**
	 * Read data files
	 * 
	 */
	static void readData() {
	    
	    System.err.println();
	    System.err.print("reading data.. ");
		
	    // pos-train
	    seqsPosTrain = ReadFasta.read(properties.getProperty("inputdatatrainingpositive"), (byte)1);

	    // neg-train	    
	    seqsNegTrain = ReadFasta.read(properties.getProperty("inputdatatrainingnegative"), (byte)0);

	    // pos-test
	    seqsPosTest = ReadFasta.read(properties.getProperty("inputdatatestpositive"), (byte)1);
	    
	    // neg-test
        seqsNegTest = ReadFasta.read(properties.getProperty("inputdatatestnegative"), (byte)0);

        System.err.print("balancing data.. ");
        // Read in again until data is balanced out
        while (seqsPosTrain.size() * 5 < seqsNegTrain.size()) {
            seqsPosTrain.addAll(ReadFasta.read(properties.getProperty("inputdatatrainingpositive"), (byte)1));
        }
        
        while (seqsNegTrain.size() * 5 < seqsPosTrain.size()) {
            seqsNegTrain.addAll(ReadFasta.read(properties.getProperty("inputdatatrainingnegative"), (byte)0));
        }
        
        System.err.println("okay");
        
        System.err.println();
        System.err.println("pos-train        :" + " " + seqsPosTrain.size());
        System.err.println("neg-train        :" + " " + seqsNegTrain.size());
        System.err.println("pos-test         :" + " " + seqsPosTest.size());
        System.err.println("neg-test         :" + " " + seqsNegTest.size());

        // Put negative and positive data together
        seqsTrain = new ArrayList<LSTMSequence>();
        
        seqsTrain.addAll(seqsPosTrain);
        seqsTrain.addAll(seqsNegTrain);
        
        seqsTest = new ArrayList<LSTMSequence>();
        
        seqsTest.addAll(seqsPosTest);
        seqsTest.addAll(seqsNegTest);
        
        System.err.println();
        System.err.println("train            :" + " " + seqsTrain.size());
        System.err.println("test             :" + " " + seqsTest.size());
		
	}
	
	/**
	 * Read only test data files
	 * 
	 */
	static void readDataTest() {
	    
	    System.err.println();
	    System.err.print("reading data for testing.. ");
		
	    // pos-test
	    seqsPosTest = ReadFasta.read(properties.getProperty("inputdatatestpositive"), (byte)1);
	    
	    // neg-test
        seqsNegTest = ReadFasta.read(properties.getProperty("inputdatatestnegative"), (byte)0);
        
        System.err.println("okay");
        
        System.err.println();
        System.err.println("pos-test         :" + " " + seqsPosTest.size());
        System.err.println("neg-test         :" + " " + seqsNegTest.size());

        // Put positive and negative data together
        
        seqsTest = new ArrayList<LSTMSequence>();
        
        seqsTest.addAll(seqsPosTest);
        seqsTest.addAll(seqsNegTest);
        
        System.err.println();
        System.err.println("test             :" + " " + seqsTest.size());
		
	}

	/**
	 * Init some stuff
	 * 
	 */
	static void init() {

        /* Number of inputs */
        in_mod = nSymbols * windowSize - 1;
          
        /* Bias unit */
        in_mod_b = in_mod + 1;
        
        /* Number of inputs for local coding */
        in_nn_mod = windowSize;
        
        cell_mod = in_mod_b;
        
        for (int i = 0; i < num_blocks; i++) {
            
          cell_mod += (2 + blockSize[i]);
        
        }
        
        cell_mod++;
        
        ges_mod = cell_mod+out_mod;
        
        System.err.println();
        int noi = in_mod + 1;
        System.err.println("Number of Blocks : " + num_blocks);
        System.err.println("Window size      : " + windowSize);
        System.err.println("Number of Inputs : " + noi);
        //System.err.println("Bias Unit        : " + in_mod_b);
        int nou = ges_mod + 1;
        System.err.println("Number of Units  : " + nou);
        System.err.println();
        if (!test) {
        	System.err.println("Learning rate    : " + alpha);
        	System.err.println();
        	System.err.println("Stop after       : " + maxepochs + " epochs");
        	System.err.println("Test every       : " + testnepochs + " epochs");
        }
        System.err.println("ROCn             : " + rocn);       
        System.err.println();
        System.err.println("Weight matrix    : " + weightfile);
        System.err.println();
        
        if (!test) {
        	if (humanreadableweightmatrix != 0) {
        		System.err.println("Additional human readable weight matrix : yes");
        	}
        	System.err.println("Writing weight matrix every             : " + writeweightsafternepochs + " epochs");      
        
        	System.err.println();
        	System.err.println("Number of Threads                       : " + numberThreads);
        	
        }
        	
        // Create weightmatrix
        wm = new WeightMatrix(ges_mod, cell_mod, in_mod, in_mod_b, num_blocks,
                              blockSize, nSymbols, init_range);
        
        wm.setWeightfile(weightfile);
        
        
		
	}
	
	/**
	 * Parse command line options
	 * 
	 * @param args
	 */
	static void parseCommandline(String[] args) {

	    if (args.length < 6 || args.length > 7) {	        
	        usage();
	    }
	    
	    if (!"-p".equals(args[0])) {
	    	usage();
	    }
	    
	    else {
	    	propertiesfile = args[1];
	    }
	    
	    if (!"-w".equals(args[2])) {	
	    	usage();
	    }
	    
	    else {	
	    	weightfile = args[3];
	    }
	    
	    if (!"-t".equals(args[4])) {	
	    	usage();
	    }
	    
	    else {	
	    	numberThreads = Integer.valueOf(args[5]);
	    }
	     
	    
	    test        = false;
	    loadweights = false;
	    
	    if (args.length == 7) {

        	if ("-test".equals(args[6])) {

        		test = true;

        	}
        	
        	else if ("-lw".equals(args[6])) {
        		
        		loadweights = true;
        		
        	}
        	
        	else {
        		
        		usage();
        		
        	}
        	
	    }
		
	}
	
	
	static void usage() {
	    
	    System.err.println();
	    System.err.println("Usage: JLSTM -p properties file -w weightfile -t threads [-lw|-test]");
	    System.exit(-1);
	    
	}
	
	/**
	 * Main
	 * 
	 * @param args
	 */
	public static void main(String[] args) {
	    
	    parseCommandline(args);
			    
	    readProperties();
	    
	    if (test) {
	    	readDataTest();
	    }
	    else {
	    	readData();
	    }

        init();
        
        /* if -lw or -test parameter set then load weights */
        
        if (test || loadweights) {
        	
        	System.err.println();

        	System.err.print("Reading weight matrix.. ");

        	wm.readWeightMatrix();

        	System.err.println("okay");
        
        	System.err.println();
        	
        }
        
        /* test with one thread */
        
        if (test) {
        	
        	LSTMThread lt = new LSTMThread(ges_mod, in_mod, in_mod_b, cell_mod, in_nn_mod,
        			num_blocks, blockSize, nSymbols, alpha,
        			bias_inp, bias_out, bias_memin,
                    weight_memout, outputbias,
                    windowSize,
        			targetpositive, targetnegative,
        			0, 0,
        			seqsPosTest.size(), seqsNegTest.size(),
        			rocn,
        			maxepochs,
        			testnepochs,
        			writeweightsafternepochs,
        			null,
        			seqsTest,
        			wm, true);
        	
        	/* test and exit */
        	lt.start();
        	
        	try {
        		lt.join();
        	}
        	catch (InterruptedException ie) {
        	}
        	
        	System.exit(0);

	    }
        
        System.err.println();
        System.err.println("Running...");
        System.err.println();
        
        // Running the threads...
        
        ArrayList<Thread> threadList = new ArrayList<Thread>();
        
        for (int thread = 0; thread < numberThreads; thread++) {
        
        	LSTMThread lt = new LSTMThread(ges_mod, in_mod, in_mod_b, cell_mod, in_nn_mod,
        			num_blocks, blockSize, nSymbols, alpha,
        			bias_inp, bias_out, bias_memin,
        			weight_memout, outputbias,
        			windowSize,
        			targetpositive, targetnegative,
        			seqsPosTrain.size(), seqsNegTrain.size(),
        			seqsPosTest.size(), seqsNegTest.size(),
        			rocn,
        			maxepochs,
        			testnepochs,
        			writeweightsafternepochs,
        			seqsTrain,
        			seqsTest,
        			wm, false);
        	
        	lt.start();
        	
        	threadList.add(lt);
        	
        	
        	/* Wait a little bit for the next thread */
            /*
        	try {
        		Thread.sleep(1000);
        	}
        	catch (InterruptedException ie){
        		// nothing 
        	}
            */
            
        }
        
        /* Wait until all threads have finished */
        for (int i = 0; i < threadList.size(); i++) {
        	
        	try {
        		threadList.get(i).join();
        	}
        	catch (InterruptedException ie) {
        		
        	}
        	
        }
        
        /* Start a last test */
        new LSTMThread(ges_mod, in_mod, in_mod_b, cell_mod, in_nn_mod,
    			num_blocks, blockSize, nSymbols, alpha,
    			bias_inp, bias_out, bias_memin,
                weight_memout, outputbias,
                windowSize,
    			targetpositive, targetnegative,
    			0, 0,
    			seqsPosTest.size(), seqsNegTest.size(),
    			rocn,
    			maxepochs,
    			testnepochs,
    			writeweightsafternepochs,
    			seqsTrain,
    			seqsTest,
    			wm, true).start();
        
	}

}
