/**
 * @author Martin Heusel
 * 
 * Copyright (C) 2008 Martin Heusel (mhe@bioinf.jku.at),
 *                      
 * Johannes Kepler University, Linz, Austria
 * Institute of Bioinformatics.
 * The software is maintained by Martin Heusel.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * If you use this library, please cite:
 *
 * @article{SeppHochreiter07152007,
 *   author = {Hochreiter, Sepp and Heusel, Martin and Obermayer, Klaus},
 *   title = {{Fast Model-based Protein Homology Detection without Alignment}},
 *   journal = {Bioinformatics},
 *   volume = {23},
 *   number = {14},
 *   pages = {1728-1736},
 *   doi = {doi:10.1093/bioinformatics/btm247},
 *   year = {2007},
 *   URL = {http://bioinformatics.oxfordjournals.org/cgi/content/abstract/btm247v1},
 *   eprint = {http://bioinformatics.oxfordjournals.org/cgi/reprint/btm247v1}
 * }
 * 
 * $Id: LSTM.java 227 2009-02-13 11:00:21Z mhe $
 *
 */


package at.jku.bioinf.jlstm;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.text.DecimalFormat;
import java.util.ArrayList;

/**
 * This is the core of the LSTM. Computes forward, backward pass and errors.
 * 
 * @author mhe
 *
 */
public class LSTM {
    
    int ges_mod;
    int in_mod;
    int in_mod_b;
    int cell_mod;
    int in_nn_mod;
    int num_blocks;
    int[] blockSize;
    
    int nSymbols;
    
    double[] yk_old;
    double[] yk_new;
    double[] y_out;
    double[] y_in;
    double[][] yc;
    double[][] g;
    double[][] h;
    double[][] s;
    double[][][] sc;
    
    double[] e;
    double[] eo;
    double[][] ec;
    double[][] es;
    
    double[][][] si;
    
    double[] error;
    double targetpositive;
    double targetnegative;
    
    double seq_err;
    double seq_correct;
    double epoch_err;
    
    int falsepos;
    int falseneg;
    
    int numpostrain;
    int numnegtrain;
    int numpostest;
    int numnegtest;
    
    int numseqstrain;
    int numseqstest;
    
    boolean ispositive;
    
    double alpha;
    
    int rocn;
    
    ArrayList<RocData> roclist;
    
    ComputeRoc cr = new ComputeRoc();
    
    double[][] inputVector;
    short[][] inputIndex;

    WeightMatrix wm;
    
    double[][] weightMatrixLocal;
    double[][] weightMatrixGlobal;
    
    double[][] DW;
    
    public LSTM (int ges_mod, int in_mod, int in_mod_b, int cell_mod, int in_nn_mod,
                 int num_blocks, int[] blockSize, int nAAs, double alpha, 
                 double targetpositive, double targetnegative,
                 int numpostrain, int numnegtrain,
                 int numpostest, int numnegtest,
                 int rocn,
                 WeightMatrix wm) {
        
        this.ges_mod        = ges_mod;
        this.in_mod         = in_mod;
        this.in_mod_b       = in_mod_b;
        this.cell_mod       = cell_mod;
        this.in_nn_mod      = in_nn_mod;
        this.num_blocks     = num_blocks;
        this.blockSize      = blockSize;
    
        this.nSymbols       = nAAs;
        
        this.numpostrain    = numpostrain;
        this.numnegtrain    = numnegtrain;
        this.numseqstrain   = numpostrain + numnegtrain;

        this.numpostest     = numpostest;
        this.numnegtest     = numnegtest;
        this.numseqstest    = numpostest + numnegtest;

        this.alpha          = alpha;
        
        this.targetpositive = targetpositive;
        this.targetnegative = targetnegative;
        
        this.rocn           = rocn;
        
        this.wm             = wm;
        
        this.weightMatrixGlobal   = wm.getWeightMatrix();
        
        // Init matrices
        
        DW     = new double[ges_mod][ges_mod];
        //DWold  = new double[ges_mod][ges_mod];
        
        yk_old = new double[ges_mod + 1];
        yk_new = new double[ges_mod + 1];
        
        y_out  = new double[num_blocks];
        y_in   = new double[num_blocks];
        yc     = new double[num_blocks][5]; // TODO: change 5 to maximum blocksize
        g      = new double[num_blocks][5];
        h      = new double[num_blocks][5];
        s      = new double[num_blocks][5];
        
        si     = new double[num_blocks][5][ges_mod + 1];
        sc     = new double[num_blocks][5][ges_mod + 1];
        
        e      = new double[ges_mod + 1];
        eo     = new double[num_blocks];
        ec     = new double[num_blocks][5];
        es     = new double[num_blocks][5];
        
        error  = new double[4];   // TODO: change 4 to number output units
        
        this.roclist = new ArrayList<RocData>(numseqstrain);
        
        this.seq_err = 0;
        this.seq_correct = 1;
        this.epoch_err = 0;
        
        this.falsepos = 0;
        this.falseneg = 0;
        
        reset();
        
    }
    
    /**
     * Fast exponential approximation
     * 
     * {@link http://martin.ankerl.com/2007/02/11/optimized-exponential-functions-for-java/}
     * 
     * @param val
     * @return
     */
    public double fastexp(double val) {  
    	final long tmp = (long) (1512775 * val + (1072693248 - 60801));  
    	return Double.longBitsToDouble(tmp << 32);  
    }  
    
    /**
     * Compute the error
     */
    void compError () {
        
        /* compute error */
        for (int i = cell_mod; i < ges_mod; i++) {
            
            error[0] = (this.ispositive) ? targetpositive - yk_new[i] : targetnegative - yk_new[i];
            
        }
        
        double err = 0;
        
        err +=  error[0] * error[0];
        
        seq_err += err;
        epoch_err += err;
        
        if ((seq_correct == 1) && (Math.abs(error[0]) > Math.abs(targetpositive - targetnegative)/2)) { 
        
          seq_correct = 0;
          
          if (this.ispositive) { 
              falseneg++;
          }
          else {
              falsepos++;
          }

        }
        
               
    }
    
    /**
     * Output of net performance (MSE, false-positives and false-negatives)
     * 
     * @param isTrain
     */
    public void printError(int epoch, boolean isTrain) {

        int numpos, numneg, numseqs;
        String what = "";
        String pre  = "";
        
        if (isTrain) {
        	numpos = numpostrain;
        	numneg = numnegtrain;
        	numseqs = numseqstrain;
        	what    = "training";
        	pre     = "";
        	
        	cr.computeAUC(roclist, numpostrain, numnegtrain, rocn);
        	
        }
        else {
        	numpos = numpostest;
        	numneg = numnegtest;
        	numseqs = numseqstest;
        	what    = "test";
        	pre     = "TEST: ";
        	
        	cr.computeAUC(roclist, numpostest, numnegtest, rocn);
        	
        }
        
        double auc = cr.getAuc();
        double aucn = cr.getAucn();
        
        double fnp = (numpos != 0) ? (double) falseneg / (double) numpos : 0;
        double fpp = (numneg != 0) ? (double) falsepos / (double) numneg : 0;
        
        double mse = epoch_err / (1.0 * numseqs);
        
        BufferedWriter out = null;

        try {
          
          out = new BufferedWriter( new FileWriter( "out.txt", true ) );
          
          out.write("epoch: " + epoch);
          out.newLine();
          out.write(pre + "MSE ");
          out.write(new DecimalFormat("#.####").format(mse));
          out.newLine();
          out.write(pre + "false-negative:");
          out.write(falseneg + " (out of " + numpos + " positive " + what + " examples) ");
          out.write(new DecimalFormat("#0.0").format(fnp * 100) + "%");
          out.newLine();
          out.write(pre + "false-positive:");
          out.write(falsepos + " (out of " + numneg + " negative " + what + " examples) ");
          out.write(new DecimalFormat("#0.0").format(fpp * 100) + "%");
          out.newLine();
          out.write(pre +  "ROC ");
          out.write(new DecimalFormat("#0.000").format(auc));
          out.newLine();
          out.write(pre +  "ROC" + rocn + " ");
          out.write(new DecimalFormat("#0.000").format(aucn));
          out.newLine();
          out.newLine();
          out.close();
          
        }
        catch ( IOException e ) {
          System.err.println( e );
        }
        finally {
          try {
            if ( out != null ) out.close();
          } catch ( IOException e ) { }
        }
        
        System.out.println("epoch: " + epoch);
        System.out.printf("%sMSE:%f\n", pre, mse);
        System.out.printf("%sfalse-negative:%d (out of %d positive %s examples) %.1f%%\n", pre, falseneg, numpos, what, fnp * 100);
        System.out.printf("%sfalse-positive:%d (out of %d negative %s examples) %.1f%%\n", pre, falsepos, numneg, what, fpp * 100);
        System.out.printf("%sROC %.3f\n", pre, auc);
        System.out.printf("%sROC" + rocn + " %.3f\n", pre, aucn);
        System.out.println();
    	
    }
    
    /**
     * Timestep
     * 
     */
    public void timeforward() {
        
        for (int i = 0; i < ges_mod; i++) {
            yk_old[i] = yk_new[i];
        }
        
        seq_err = 0;
        seq_correct = 1;
        
    }
    
    public void cloneWeightMatrix() {
    	
    	this.weightMatrixLocal = this.weightMatrixGlobal.clone();
    	
    }
    
    /**
     * Forward pass
     * 
     * @param element
     * @param targ
     */
    public void forwardPass(int element, boolean targ) {
    	
        /* ### memory cells ### */

        int i = in_mod_b + 1;
        
        for (int u = 0; u < num_blocks; u++) {

            double sum;
            
            /* input gates */
            
            sum = 0;
            
            /* input */
            /*
            for (int j=0;j<in_nn_mod;j++) {
                sum += weightMatrix[i][inputIndex[element][j]] * inputVector[element][j];
            }
            */
            
            /* lstm */
            for (int j = in_mod_b; j < cell_mod; j++) {
                sum += weightMatrixLocal[i][j] * yk_old[j];
            }

            y_in[u] = 1/(1+fastexp(-sum));
            yk_new[i]= y_in[u];

            /* output gates */
            
            i++;
            
            /* input */
            
            sum = 0;
            
            /*
            for (int j=0;j<in_nn_mod;j++) {
                sum += weightMatrixLocal[i][inputIndex[element][j]] * inputVector[element][j];
            }
            */
            
            /* lstm */
            for (int j = in_mod_b; j < cell_mod; j++) {
                sum += weightMatrixLocal[i][j] * yk_old[j];
            }

            /* peep hole */
            //sum -= S[u][0];
            
            y_out[u]  = 1/(1+fastexp(-sum));
            
            yk_new[i] = y_out[u];

            /* uth memory cell block */
            for (int v = 0; v < blockSize[u]; v++) {
                
                /* activation of function g of vth memory cell of block u  */
                i++;
                sum = weightMatrixLocal[i][in_mod_b]; // bias

                for (int j = 0; j < in_nn_mod; j++) {
                    sum += weightMatrixLocal[i][inputIndex[element][j]] * inputVector[element][j];
                }

                /* peep hole */
                //sum -= S[u][v];
                //sumblock[u] = sum;

                g[u][v] = 2.0/(1+fastexp(-sum));
                
                /* update internal state  */
                s[u][v] += y_in[u]*g[u][v];
                
                /* activation function h */
                if (s[u][v] > 500.0) {
                    //System.err.println("S drifted\n");
                    h[u][v] = 1.0;
                }
                else {
                    h[u][v] = 2.0/(1+fastexp(-s[u][v]))-1.0;
                }

                /* activation of vth memory cell of block u */
                yc[u][v] = h[u][v]*y_out[u];
                yk_new[i] = yc[u][v];
                
                
                
            }
            
            i++;
            
        }



        /* ### output units activation ### */

        if (targ) /* only if target for this input */ {

            for (int k = cell_mod; k < ges_mod; k++) {        
                
                /* bias */
                double sum = weightMatrixLocal[k][in_mod_b];
                
                /* memory cells input */
                i = in_mod_b + 1;
                for (int u = 0; u < num_blocks; u++) {
                    i++;
                    for (int v = 0; v < blockSize[u]; v++) {
                        i++;
                        sum += weightMatrixLocal[k][i] * yk_new[i];
                    }
                    i++;
                }
                
                /* activation */
                yk_new[k] = 1/(1+fastexp(-sum));
                
                //System.out.println(yk_new[k]);
                
                /* Store output and label in RocData container */
                byte label = (byte) ((ispositive) ? 1 : 0);
                RocData rd = new RocData(yk_new[k], label);
                
                roclist.add(rd);
                

            }       

        }

        
    }
    
    
    /**
     * Backward pass
     * 
     * @param element
     */
    public void backwardPass(int element) {
        
        /* output units */
        for (int k = cell_mod, j = 0; k < ges_mod; k++, j++) {
            
            e[k] = error[j] * (1.0 - yk_new[k]) * yk_new[k]; // log
            
            /* weight update contribution */
            
            double tmp = alpha * e[k];
               
            int i = in_mod_b + 1;
            
            for (int u = 0; u < num_blocks; u++) {
                
                i++;
                
                for (int v = 0; v < blockSize[u]; v++) {
                    
                    i++;
                    
                    DW[k][i] += tmp*yk_new[i];
                    
                }
                
                i++;
                
            }
            
            /* bias to output unit */
            DW[k][in_mod_b] += tmp;
            
        }
        
        /* error to memory cells ec[][] and internal states es[][] */

        int i = in_mod_b + 1;
        
        /* input gates */
        for (int u = 0; u < num_blocks; u++) {
            
            i++;
            
            for (int v = 0; v < blockSize[u]; v++) {
                
                i++;
                
                double sum = 0;
                
                for (int k = cell_mod; k < ges_mod; k++) {
                    
                    sum += weightMatrixLocal[k][i] * e[k];
                    
                }
                
                ec[u][v] = sum;
                
                es[u][v] = 0.5 * y_out[u] * (1.0 + h[u][v]) * (1.0 - h[u][v]) * sum;
                
            }
            
            i++;
            
        }

        /* output gates */
        for (int u = 0; u < num_blocks; u++) {
            
            double sum = 0;
            
            for (int v = 0;v < blockSize[u]; v++) {
                
                sum+= h[u][v] * ec[u][v];
                
            }
            
            eo[u]=sum * (1.0 - y_out[u]) * y_out[u];
            
        }

        /* Derivatives of the internal state */

        derivatives(element);


        /* updates for weights to input and output gates and memory cells */
        
        i = in_mod_b + 1;
        
        for (int u = 0; u < num_blocks; u++) {        
            
            /* input gates */
            
            /* input */
        	/*
            for (int j=0;j<in_mod_b;j++) {
                double sum = 0;
                for (int v=0;v<blockSize[u];v++) {
                    sum += es[u][v]*si[u][v][j];  
                }
                DW[i][j] += sum;
            }
            */
            
            /* lstm */
            for (int j = in_mod_b; j < cell_mod; j++) {
                
                double sum = 0;
                
                for (int v = 0; v < blockSize[u]; v++) {
                    
                    sum += es[u][v] * si[u][v][j];
                    
                }
                
                DW[i][j] += alpha * sum;
                
            }
            
            /*output gates */
            
            i++;
            
            double tmp = alpha * eo[u];
            
            /* input */
            /*
            for (int j=0;j<in_mod;j++) {
                DW[i][j] += tmp * yk_old[j];
            }
            */
            
            /* lstm */
            for (int j = in_mod_b; j < cell_mod; j++) {
                
                DW[i][j] += tmp * yk_old[j];
                
            }
            
            /* memory states */
            for (int v = 0; v < blockSize[u]; v++) {
                
                i++;
                
                tmp = alpha * es[u][v];
                
                /* input */
                for (int j = 0; j <= in_mod_b; j++) {
                    
                    DW[i][j] += tmp*sc[u][v][j];
                    
                }
                
                /* lstm & bias */
                /*
                for (int j=in_mod_b;j<cell_mod;j++) {
                    DW[i][j] += tmp*sc[u][v][j];           
                }
                */
                
            }
            
            i++;
            
        }

        
    }
    
    /**
     * Computes derivatives
     * 
     * @param element
     */
    public void derivatives(int element) {
              
        for (int u = 0; u < num_blocks; u++) {
            
            //tmp0 = (1.0-Y_in[u])*Y_in[u];
            //tmp1 = Y_in[u]*2.0;
            
            for (int v = 0; v < blockSize[u]; v++) {
                
                double tmp1 = y_in[u] * g[u][v];
                
                double tmp2 = (1.0 - y_in[u]) * tmp1;
                
                double tmp3 = (1.0 - 0.5 * g[u][v]) * tmp1;
                
                for (int j = 0; j < in_nn_mod; j++) {
                    
                    /* weights to input gate */
                    //si[u][v][inputIndex[element][j]] += tmp2 * inputVector[element][j];
                    
                    /* weights to cell input */
                    sc[u][v][inputIndex[element][j]] += tmp3 * inputVector[element][j];
                
                }
                
                for (int j = in_mod_b; j < cell_mod; j++) {
                    
                    /* weights to input gate */
                    si[u][v][j] += tmp2 * yk_old[j];
                    
                    /* weights to cell input */
                    //sc[u][v][j] += tmp3 * yk_old[j];
                    
                }
                
                sc[u][v][in_mod_b] += tmp3;
                
                //sc[u][v][in_mod_b] = 0;
            }
            
        }
        
    }
    
      
    /**
     * Initialization of the net
     * 
     * @param bias_inp
     * @param bias_out
     * @param bias_memin
     * @param weight_memout
     * @param outputbias
     */
    public void init (double[] bias_inp, double[] bias_out,
                      double[] bias_memin, double[] weight_memout,
                      double outputbias) {
        
    	cloneWeightMatrix();
    	
        /* bias and weight to output unit initalizations */
        int i = in_mod_b + 1;
        for (int u = 0; u < num_blocks; u++) {
            /* bias to input gate */
            weightMatrixLocal[i][in_mod_b] = bias_inp[u];
            i++;
            /* bias to output gate */
            weightMatrixLocal[i][in_mod_b] = bias_out[u];
            for (int v = 0; v < blockSize[u]; v++) {
                i++;
            }
            /* bias to memory cell */
            weightMatrixLocal[i][in_mod_b] = bias_memin[u];
            /* weight to output unit */ 
            weightMatrixLocal[cell_mod][i] = weight_memout[u];

            i++;
        }

        /* output bias */
        weightMatrixLocal[cell_mod][in_mod_b] = outputbias;

        
    }
    
    /**
     * Reset the net
     * 
     */
    public void reset () {
        
        for (int i = 0; i < ges_mod; i++) {
            
            yk_new[i] = 0.5;
            yk_old[i] = 0.5;
            
            for (int j = 0; j < ges_mod; j++) {
            	DW[i][j] = 0;
            }
            
        }
        
        yk_new[in_mod_b] = 1.0;
        yk_old[in_mod_b] = 1.0;
        
        for (int u = 0; u < num_blocks; u++) {
            
            y_in[u]  = 0;
            y_out[u] = 0;
            
            for (int v = 0; v < blockSize[u]; v++) {
                
                s[u][v]  = 0;
                g[u][v]  = 0;
                h[u][v]  = 0;
                yc[u][v] = 0.5;
                
                for (int j = 0; j < cell_mod; j++) {
                    
                    si[u][v][j] = 0;
                    sc[u][v][j] = 0;
                    
                }
                
           }
            
        }

    }
    
    public void setInputVector(double[][] inputVector) {
        this.inputVector = inputVector;
    }

    public void setInputIndex(short[][] inputIndex) {
        this.inputIndex = inputIndex;
    }


    public void setIsPositive (boolean ispositive) {
        this.ispositive = ispositive;
    }

	public void setEpoch_err(double epoch_err) {
		this.epoch_err = epoch_err;
	}

	public void setFalsepos(int falsepos) {
		this.falsepos = falsepos;
	}

	public void setFalseneg(int falseneg) {
		this.falseneg = falseneg;
	}
	
	public ArrayList<RocData> getRoclist() {
        return roclist;
    } 
    
	public double[][] getDW() {
		return DW;
	}
	
}
