/**
 * @author Martin Heusel
 * 
 * Copyright (C) 2008 Martin Heusel (mhe@bioinf.jku.at),
 *                      
 * Johannes Kepler University, Linz, Austria
 * Institute of Bioinformatics.
 * The software is maintained by Martin Heusel.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * If you use this library, please cite:
 *
 * @article{SeppHochreiter07152007,
 *   author = {Hochreiter, Sepp and Heusel, Martin and Obermayer, Klaus},
 *   title = {{Fast Model-based Protein Homology Detection without Alignment}},
 *   journal = {Bioinformatics},
 *   volume = {23},
 *   number = {14},
 *   pages = {1728-1736},
 *   doi = {doi:10.1093/bioinformatics/btm247},
 *   year = {2007},
 *   URL = {http://bioinformatics.oxfordjournals.org/cgi/content/abstract/btm247v1},
 *   eprint = {http://bioinformatics.oxfordjournals.org/cgi/reprint/btm247v1}
 * }
 * 
 * $Id: ComputeRoc.java 102 2008-10-08 14:46:53Z mhe $
 *
 */


package at.jku.bioinf.jlstm;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Random;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;


/**
 * LSTM Thread for running a thread
 * Has it's own lstm core and sequence data containers
 * Weight matrix is shared across all threads only the weight update
 * is synchronized.
 * 
 * @author mhe
 *
 */
public class LSTMThread extends Thread {
	
	ArrayList<LSTMSequence> seqsTrain;
	ArrayList<LSTMSequence> seqsTest;
	
	int numpostrain;
    int numnegtrain;
    int numpostest;
    int numnegtest;

	/* Input matrix */
    InputMatrix im;
    
    /* LSTM core */
    LSTM lstm;
    
    /* Weight matrix */
    WeightMatrix wm;
	
    /* window size */
    int windowSize;
	
	/* Number of symbols. 20 or 23 for Proteins and 4 for DNA. */
    int nSymbols;
    
    /* Index of output unit */
    int out_mod;
    
    /* Number of LSTM blocks */
    int num_blocks;
    
    /* Index of last input unit */
    int in_mod;
    
    /* Bias unit */
    int in_mod_b;
    
    /* Number of inputs for local coding */
    int in_nn_mod;
    
    /* Index of last LSTM unit */
    int cell_mod;
    
    /* Number of units */
    int ges_mod;

    /* Blocksize for every block */
    int[] blockSize;
    
    /* Range for random init of weights */
    double init_range;
    
    /* Target values */
    double targetpositive;
    double targetnegative;

    /* Learning rate */
    double alpha;
    
    /* Biases and initial weights */
    double[] bias_inp;
    double[] bias_out;
    double[] bias_memin;
    double[] weight_memout;
    double outputbias;
    
    /* Maximum number of epochs before stop */
    int maxepochs;
    
    /* After testnepochs a test is performed */
    int testnepochs;
    
    /* After writeweightsafternepochs the weight matrix is written */
    int writeweightsafternepochs;
    
    /* Write an additional better readable weight matrix? For debugging
     * and investigation.
     */
    int humanreadableweightmatrix;
    
    boolean test;
    
    /* Additional ROCN (e.g. N = 50) to be computed */
    int rocn;
    
    /* epoch is shared across all threads */
    static int epoch;
    
    ReentrantReadWriteLock rwl = new ReentrantReadWriteLock();
    private final Lock wlock = rwl.writeLock();
    
    public LSTMThread(int ges_mod, int in_mod, int in_mod_b, int cell_mod, int in_nn_mod,
            int num_blocks, int[] blockSize, int nAAs, double alpha,
            double[] bias_inp, double[] bias_out, double[] bias_memin,
            double[] weight_memout, double outputbias,
            int windowSize,
            double targetpositive, double targetnegative,
            int numpostrain, int numnegtrain,
            int numpostest, int numnegtest,
            int rocn,
            int maxepochs,
            int testnepochs,
            int writeweightsafternepochs,
            ArrayList<LSTMSequence> seqsTrain,
            ArrayList<LSTMSequence> seqsTest,
            WeightMatrix wm, boolean test) {
    	
    	this.ges_mod        = ges_mod;
        this.in_mod         = in_mod;
        this.in_mod_b       = in_mod_b;
        this.cell_mod       = cell_mod;
        this.in_nn_mod      = in_nn_mod;
        this.num_blocks     = num_blocks;
        this.blockSize      = blockSize;
        
        this.bias_inp       = bias_inp;
        this.bias_out       = bias_out;
        this.bias_memin     = bias_memin;
        this.weight_memout  = weight_memout;
        this.outputbias     = outputbias;
        
        this.windowSize     = windowSize;
    
        this.nSymbols       = nAAs;
        
        this.numpostrain    = numpostrain;
        this.numnegtrain    = numnegtrain;
        

        this.numpostest     = numpostest;
        this.numnegtest     = numnegtest;
        
        this.alpha          = alpha;
        
        this.targetpositive = targetpositive;
        this.targetnegative = targetnegative;
        
        this.rocn           = rocn;
        
        this.maxepochs      = maxepochs;
        
        this.testnepochs    = testnepochs;
        
        this.writeweightsafternepochs = writeweightsafternepochs;
        
        this.test           = test;
        
        if (!test) {
        
        	this.seqsTrain      = (ArrayList<LSTMSequence>)seqsTrain.clone();
        	
        }
        
        this.seqsTest       = (ArrayList<LSTMSequence>)seqsTest.clone();
        
        this.wm             = wm;
        
    	// Create inputmatrix
        im = new InputMatrix();
        
        // Make LSTM network
        if (test) {
        	
        	lstm = new LSTM(ges_mod, in_mod, in_mod_b, cell_mod, in_nn_mod,
                     num_blocks, blockSize, nSymbols, alpha,
                     targetpositive, targetnegative,
                     0, 0,
                     numpostest, numnegtest,
                     rocn,
                     wm);
        	
        }
        else {
        	
        	lstm = new LSTM(ges_mod, in_mod, in_mod_b, cell_mod, in_nn_mod,
        			num_blocks, blockSize, nSymbols, alpha,
        			targetpositive, targetnegative,
        			numpostrain, numnegtrain,
        			numpostest, numnegtest,
        			rocn,
        			wm);
        	
        	lstm.init (bias_inp, bias_out,bias_memin, weight_memout, outputbias);
        	
        }
        	
        
    	
    }
	
	/**
	 * Training
	 * 
	 * @param epoch
	 */
	void training(int epoch) {
		
		// For all training sequences
        for (int i = 0; i < seqsTrain.size(); i++) {
        	
            // Create input matrix
            im.computeInputs(seqsTrain.get(i), windowSize, nSymbols);
            
            // Set  input index and input vector in the lstm core
            lstm.setInputIndex(im.getInputIndex());
            lstm.setInputVector(im.getInputVector());
            
            // Set label 
            lstm.setIsPositive (seqsTrain.get(i).label == 1);
            
            // For all elements in the current sequence i except the last one
            int j;
            for (j = 0; j < seqsTrain.get(i).getLocalCodings().length - 1; j++) {
            	lstm.cloneWeightMatrix();
                lstm.forwardPass(j, false);
                lstm.derivatives(j);
                lstm.timeforward();
            }
            
            // The last element has a target
            lstm.cloneWeightMatrix();
            lstm.forwardPass(j, true);
            lstm.compError();
            lstm.backwardPass(j);
            
            /* write lock the weight update */ 
            wlock.lock();
            wm.weightUpdate(lstm.getDW());
            wlock.unlock();
            
            lstm.reset();          
        
        }
        
        lstm.printError(epoch, true);
        lstm.setEpoch_err(0);
        lstm.setFalsepos(0);
        lstm.setFalseneg(0);		
		
		
	}
	
	/**
	 * Test
	 * 
	 * @param epoch
	 */
	void test(int epoch) {
		
    	// For all test sequences
    	for (int i = 0; i < seqsTest.size(); i++) {

    		im.computeInputs(seqsTest.get(i), windowSize, nSymbols);
    		
    		lstm.setInputIndex(im.getInputIndex());
            lstm.setInputVector(im.getInputVector());
            
    		
    		lstm.setIsPositive(seqsTest.get(i).label == 1);
    		
    		lstm.cloneWeightMatrix();
    		
    		int j;
    		
    		for (j = 0; j < seqsTest.get(i).getLocalCodings().length - 1; j++) {
    			lstm.forwardPass(j, false);
    			lstm.timeforward();
    		}
    		
    		lstm.forwardPass(j, true);
    		lstm.compError();
    		lstm.reset();           

    	}
    	
    	lstm.printError(epoch, false);
    	lstm.setEpoch_err(0);
    	lstm.setFalsepos(0);
    	lstm.setFalseneg(0);
		
		
	}	
	
	
	/* (non-Javadoc)
	 * @see java.lang.Thread#run()
	 */
	@Override public void run() {
		
		int myepoch;
		
		if (test) {
			
			test(epoch);
			
			return;
			
		}
		
		Lock lock = new ReentrantLock();
		
		while (epoch <= maxepochs - 1) {
			
			lock.lock();
            
            epoch++;
            myepoch = epoch;
            
            lock.unlock();
            
            // Shuffle data set
            Collections.shuffle(seqsTrain, new Random());
            
            // Training
            
            training(myepoch);
            
            // Test after n epochs
            
            if ((myepoch % testnepochs == 0 && myepoch > 0) || myepoch == maxepochs) {
            
            	test(myepoch);
            
            }
            
            // Write weight matrix
            
            if ((myepoch % writeweightsafternepochs == 0 && myepoch > 0) || myepoch == maxepochs) {
                
                wm.writeWeightMatrix();
                
                if (humanreadableweightmatrix != 0) {
                    wm.writeReadableWeightMatrix();
                }
                
            }

        }
		
	}
	

}
