/**
 * 
 * Copyright (C) 2008 Martin Heusel (mhe@bioinf.jku.at),
 *                      
 * Johannes Kepler University, Linz, Austria
 * Institute of Bioinformatics.
 * The software is maintained by Martin Heusel.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * If you use this library, please cite:
 *
 * @article{SeppHochreiter07152007,
 *   author = {Hochreiter, Sepp and Heusel, Martin and Obermayer, Klaus},
 *	 title = {{Fast Model-based Protein Homology Detection without Alignment}},
 *	 journal = {Bioinformatics},
 *	 volume = {23},
 *	 number = {14},
 *	 pages = {1728-1736},
 *	 doi = {doi:10.1093/bioinformatics/btm247},
 *	 year = {2007},
 *	 URL = {http://bioinformatics.oxfordjournals.org/cgi/content/abstract/btm247v1},
 *	 eprint = {http://bioinformatics.oxfordjournals.org/cgi/reprint/btm247v1}
 * }
 * 
 * $Id: ReadFasta.java 236 2009-02-17 18:32:40Z mhe $
 *
 */

package at.jku.bioinf.jlstm;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.NoSuchElementException;

import org.biojava.bio.BioException;
import org.biojavax.bio.seq.RichSequence;
import org.biojavax.bio.seq.RichSequenceIterator;
import org.biojavax.bio.seq.RichSequence.IOTools;

/**
 * Reads a FASTA file and computes indices for sparse local coding.
 * 
 * @author mhe
 */
public class ReadFasta {
    
    
 
    /* 
     * Returns the position of given char in aaChars array
     * 
     */
    static byte getAACharPos(char aa) {

        /*
          ALA ARG ASN ASP CYS GLN GLU GLY HIS ILE LEU LYS MET PHE PRO SER THR 
          TRP TYR VAL ASX GLX X
        */

        final char aaChars[] =
            {'A', 'R', 'N', 'D', 'C', 'Q', 'E', 'G', 'H', 'I', 'L', 'K', 'M',
             'F', 'P', 'S', 'T', 'W', 'Y', 'V', 'B', 'Z', 'X'}; 


        int i;
        for (i = 0; i < 23; i++) {
            if (aaChars[i] == aa) {
                return (byte)i;
            }
        }
        if (aa == 'x' || aa == 'U') {
            return 22;
        }
        
        /* TODO throw Exception */
        System.out.println("Error in sequencegeneration, unknown AA Symbol");
        System.out.println("AA Symbol: " + aa);
        System.exit(-1);
        
        return 0;

    }




    /**
     * Reads a FASTA file with help of Biojava RichSequence.IOTools.
     * Creates a LSTMSequence for any sequence
     * and label and stores it in an ArrayList which is returned. 
     * 
     * @param filename
     * @param label
     * @return ArrayList<LSTMSequence>
     */
    public static ArrayList<LSTMSequence> read (String filename, byte label) {
        
        ArrayList<LSTMSequence> seqs = new ArrayList<LSTMSequence>();
        
        try {
            // Setup file input
            BufferedReader br =
                new BufferedReader(new FileReader(filename));
            
            // RichSequenceIterator
            RichSequenceIterator rsi = IOTools.readFastaProtein(br, null);
            
            // For each sequence do
            while(rsi.hasNext()) {
            	
                try {
                    
                    RichSequence rs = rsi.nextRichSequence();
                    
                    //System.out.println(rs.getName() + " " + rs.getDescription());
                    
                    // Sequence data
                    char seq[] = new char[rs.length()];
                    short localCodings[] = new short[rs.length()];
                    
                    rs.seqString().getChars(0, rs.length(), seq, 0);
                    
                    // Compute local coding indices
                    for (int i = 0; i < rs.length(); i++) {
                        localCodings[i] = getAACharPos(seq[i]);
                    }
                    
                    // Store codings and label in a LSTMSequence
                    LSTMSequence lstmseq = new LSTMSequence(localCodings, label);
                    
                    // Add LSTMSequence to ArrayList
                    seqs.add(lstmseq);
                    
                }
                catch (BioException ex) {
                    //not in fasta format or wrong alphabet
                    System.out.println("not in fasta format or wrong alphabet");
                    System.exit(0);
                }
                catch (NoSuchElementException ex) {
                    //no fasta sequences in the file
                    System.out.println("no fasta sequences in the file");
                    System.exit(0);
                }
                
            }
            
            br.close();
            
        }
        catch (FileNotFoundException fnfex) {
            //problem reading file
            System.err.println("jLSTM: Could not open sequence file: " + filename);
            System.exit(1);
        }
        
        catch (IOException ioe) {
        	  //problem reading file
            System.err.println("jLSTM: Could not close sequence file: " + filename);
            //System.exit(1);
        }
        
        return seqs;
        
    }
}
