/**
 * @author Martin Heusel
 * 
 * Copyright (C) 2008 Martin Heusel (mhe@bioinf.jku.at),
 *                      
 * Johannes Kepler University, Linz, Austria
 * Institute of Bioinformatics.
 * The software is maintained by Martin Heusel.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * If you use this library, please cite:
 *
 * @article{SeppHochreiter07152007,
 *   author = {Hochreiter, Sepp and Heusel, Martin and Obermayer, Klaus},
 *   title = {{Fast Model-based Protein Homology Detection without Alignment}},
 *   journal = {Bioinformatics},
 *   volume = {23},
 *   number = {14},
 *   pages = {1728-1736},
 *   doi = {doi:10.1093/bioinformatics/btm247},
 *   year = {2007},
 *   URL = {http://bioinformatics.oxfordjournals.org/cgi/content/abstract/btm247v1},
 *   eprint = {http://bioinformatics.oxfordjournals.org/cgi/reprint/btm247v1}
 * }
 * 
 * $Id: WeightMatrix.java 232 2009-02-16 07:30:13Z mhe $
 *
 */

package at.jku.bioinf.jlstm;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.text.DecimalFormat;
import java.util.InputMismatchException;
import java.util.Random;
import java.util.Scanner;

/**
 * Represents a weight matrix with some methods.
 * 
 * @author mhe
 *
 */
public class WeightMatrix {
    
    double[][] weightMatrix;
    
    int ges_mod;
    int in_mod;
    int in_mod_b;
    int cell_mod;
    int num_blocks;
    int nSymbols;
    
    int[] block_size;
    
    
    int wmdimx;
    int wmdimy;
    int wmdimyoffset;
    
    String weightfile;
    
    public WeightMatrix (int ges_mod, int cell_mod, int in_mod, int in_mod_b,
                         int num_blocks, int[] block_size, int nSymbols, double init_range) {
        
        //final int randfactor = 1000000;
        
        this.ges_mod    = ges_mod;
        this.cell_mod   = cell_mod;
        this.in_mod     = in_mod;
        this.in_mod_b   = in_mod_b;
        this.num_blocks = num_blocks;
        this.block_size = block_size;
        this.nSymbols   = nSymbols;
        
        weightMatrix = new double[ges_mod][ges_mod];
    
        wmdimx       = ges_mod;
        wmdimyoffset = in_mod_b + 1;
        wmdimy       = ges_mod - wmdimyoffset;
        
        
        // Init (all weights are set to zero here)
        for (int i = 0; i < ges_mod; i++) {
            for (int j = 0;j <= in_mod; j++) {
                
            	/*
                weightMatrix[i][j] = seprand(randfactor * 2) - randfactor;
                weightMatrix[i][j] /= randfactor;
                weightMatrix[i][j]*=init_range;
                */
            	
                weightMatrix[i][j] = 0;

            }
            for (int j = in_mod_b;j < ges_mod; j++) {
                
            	/*
                weightMatrix[i][j] = seprand(randfactor * 2) - randfactor;
                weightMatrix[i][j] /= randfactor;
                weightMatrix[i][j]*=init_range;
                */
            	
                weightMatrix[i][j] = 0;
                
            }
        }
        
        
    }
    
    /**
     * Reads a weight matrix from file system
     * 
     */
    public void readWeightMatrix() {
    	
    	/* reset weight matrix to zero */
    	for (int i = 0; i < this.wmdimx; i++) {
    		for (int j = 0; j < this.wmdimx; j++) {
    			weightMatrix[i][j] = 0;
    		}
    	}
    	
    	Scanner s = null;
    	
    	try {
        	
            s = new Scanner(new BufferedReader( new FileReader( weightfile ) ));

            String dim = s.next("dim");
            
            if (!"dim".equals(dim)) {
            	
            	System.err.println("Error in weight matrix. Wrong dimension header");
            	System.exit(100);
            	
            }
            
            this.wmdimy = s.nextInt();
            this.wmdimx = s.nextInt();
            this.wmdimyoffset = this.wmdimx - this.wmdimy;

            for ( int i = 0; i < wmdimy; i++ ) {
          	  
                int j;
                
                int k = wmdimyoffset + i;
                
                for (j = 0; j < wmdimx; j++) {
              	  
                    weightMatrix[k][j] = s.nextDouble();
                
                }
                
            }
  
            s.close();
            
          }
          
          catch ( FileNotFoundException fnfe ) {
        	  System.err.println( "Weight matrix file not found" );
        	  System.exit(101);
          }
          
          catch (InputMismatchException ime ) {
        	  
        	  System.err.println( "Error in weight matrix, input mismatch" );
        	  System.exit(102);
        	  
          }
          
          finally {
            
        	  s.close();
        	  
          }
        
    }
    
    /**
     * writes the current weight matrix to a file
     * 
     */
    public void writeWeightMatrix() {
        
        BufferedWriter out = null;

        try {
        	
          out = new BufferedWriter( new FileWriter( weightfile ) );
          
          out.write("dim " + String.valueOf(wmdimy) + " " + String.valueOf(wmdimx) + "\n");

          for ( int i = 0; i < wmdimy; i++ ) {
        	  
              int j;
              
              int k = wmdimyoffset + i;
              
              for (j = 0; j < wmdimx - 1; j++) {
            	  
                  out.write(new DecimalFormat("#.####").format(weightMatrix[k][j]) + " ");
              
              }
              
              out.write(new DecimalFormat("#.####").format(weightMatrix[k][j]));
              
              out.write("\n");
              
          }
          
          out.write("\n");
          
          out.close();
          
        }
        
        catch ( IOException e ) {
          System.err.println( e );
        }
        
        finally {
          try {
            if ( out != null ) out.close();
          } catch ( IOException e ) { }
        }
        
    }
    
    
    
    /**
     * write a better readable weight matrix
     * 
     */
    public void writeReadableWeightMatrix() {
        
        // TODO: Better formatting
        
        BufferedWriter out = null;
        
        String rweightfile = "readable_" + weightfile;

        try {
          
          out = new BufferedWriter( new FileWriter( rweightfile ) );
          
          out.write("dim " + String.valueOf(wmdimy) + ":" + String.valueOf(wmdimx));
          out.newLine();

          for ( int i = 0; i < wmdimy; i++ ) {
              int j;
              int k = wmdimyoffset + i;
              for (j = 0; j < wmdimx - 1; j++) {
                  out.write("(" + String.valueOf(k) + "," + String.valueOf(j) + "): ");
                  out.write(new DecimalFormat("#.####").format(weightMatrix[k][j]) + " ");
              }
              out.write("(" + String.valueOf(k) + "," + String.valueOf(j) + "): ");
              out.write(new DecimalFormat("#.####").format(weightMatrix[k][j]));
              out.newLine();
          }
          out.newLine();
          out.close();
        }
        catch ( IOException e ) {
          System.err.println( e );
        }
        finally {
          try {
            if ( out != null ) out.close();
          } catch ( IOException e ) { }
        }
        
    }
    
 
    /**
     * Computes weight updates
     * 
     * A regularization is done on the input weights with the Gonnet500 matrix
     * 
     * @param DW
     */
    public void weightUpdate(double[][] DW) {
        
        double[] dwTmp = new double[nSymbols];
        
        //final double wd = 0.999999;
        
        /* output units */

        for (int k = cell_mod, j = 0; k < ges_mod; k++, j++) {
            
            /* memory cell */
            
            int i = in_mod_b + 1;
            
            for (int u = 0; u < num_blocks; u++) {
                
                i++;
                
                for (int v = 0; v < block_size[u]; v++) {
                   
                    i++;
                    
                    weightMatrix[k][i] +=  DW[k][i];
                    
                }
                
                i++;
                
            }
            
            /* bias */
            
            weightMatrix[k][in_mod_b] +=  DW[k][in_mod_b];
            
            
        }

        /* memory cells with gates */
        
        int i = in_mod_b + 1;
        
        for (int u = 0; u < num_blocks; u++) {

        	/* input gate */
            for (int j = in_mod_b; j < cell_mod; j++) {
                
                weightMatrix[i][j] +=  DW[i][j];
                
            }
           
            i++;

            /* output gate */
            for (int j = in_mod_b; j < cell_mod; j++) {
                
                weightMatrix[i][j] += DW[i][j];
                
            }
            
            /* memory cell */
            for (int v = 0; v < block_size[u]; v++) {
                
                i++;          
                
                
                /* Regularization with the Gonnet500 substitution matrix */
                
                int j0 = 0; // begin of current aa segment
                int k0 = 0; // index of current aa in current aa segment (0-19)
                
                
                for (int k = 0; k < nSymbols; k++) {
                    
                    dwTmp[k] = 0;
                
                } 
                   
                for (int j = 0; j < in_mod_b; j++, k0++) {
                        
                        if (k0 == nSymbols) {
                          
                          for (int k = 0; k < nSymbols; k++) {
                              
                              weightMatrix[i][j0 + k] +=  dwTmp[k];
                              
                              //weightMatrix[i][j0 + k] *= wd;
                              
                              dwTmp[k] = 0;
                              
                          }
                          
                          j0 += nSymbols;
                          
                          k0 = 0;
                          
                        }
                        
                        for (int k = 0; k < nSymbols; k++) {
                            
                            // Gonnet500 matrix gives good results                        
                            if (RegularizationMatrices.getGonnet500skal0()[k0][k] != 0) {
                                
                                dwTmp[k] += DW[i][j0 + k] * RegularizationMatrices.getGonnet500skal0()[k0][k];
                            
                            }
                            
                        }   
                        
                } 
                
                /* Regularization only on the input weights */
                for (int k = 0; k < nSymbols; k++) {
                    
                    weightMatrix[i][j0 + k] +=  dwTmp[k];
                    
                    //weightMatrix[i][j0 + k] *= wd;
                
                }
                
                /* Recurrent connections from other memory cells */
                for (int j=in_mod_b;j<cell_mod;j++) {
                    
                    weightMatrix[i][j] +=  DW[i][j];
                    
                }            
                
            }
            
            i++;
            
        }

        
    }

    
    
    /**
     * set the weight file name
     * 
     * @param weightfile
     */
    public void setWeightfile(String weightfile) {
        this.weightfile = weightfile;
    }

    /**
     * get the weight matrix
     * 
     * @return weightMatrix
     */
    public double[][] getWeightMatrix() {
        return weightMatrix;
    }

    /**
     * Computes a random integer
     * 
     * @param k
     * @return
     */
    int seprand(int k) {
        long l;
        int f;
        Random rand = new Random();
        l = rand.nextLong();
        f = (int)l % k;
        return(f);
      }


}
