/**
 * @author Martin Heusel
 * 
 * Copyright (C) 2008 Martin Heusel (mhe@bioinf.jku.at),
 *                      
 * Johannes Kepler University, Linz, Austria
 * Institute of Bioinformatics.
 * The software is maintained by Martin Heusel.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * If you use this library, please cite:
 *
 * @article{SeppHochreiter07152007,
 *   author = {Hochreiter, Sepp and Heusel, Martin and Obermayer, Klaus},
 *   title = {{Fast Model-based Protein Homology Detection without Alignment}},
 *   journal = {Bioinformatics},
 *   volume = {23},
 *   number = {14},
 *   pages = {1728-1736},
 *   doi = {doi:10.1093/bioinformatics/btm247},
 *   year = {2007},
 *   URL = {http://bioinformatics.oxfordjournals.org/cgi/content/abstract/btm247v1},
 *   eprint = {http://bioinformatics.oxfordjournals.org/cgi/reprint/btm247v1}
 * }
 * 
 * $Id: InputMatrix.java 301 2009-04-21 12:20:09Z mhe $
 *
 */

package at.jku.bioinf.jlstmlogregression;


/**
 * Holds the input data for LSTM network. For a sequence the
 * complete input matrix is created.
 * 
 * @author mhe
 *
 */
public class InputMatrix {
    
    double[][]  inputVector;
    short[][] inputIndex;
    
    int windowSize;
    int nSymbols;
    int kmer;
    int numInputs;
    int sequenceCodingLength;
    int numGaussians;
    
    public InputMatrix(int windowSize, int nSymbols, int kmer, int numGaussians, int maxSeqLen) {
    	
    	this.windowSize = windowSize;
    	this.nSymbols   = nSymbols;
    	this.kmer       = kmer;
    	this.numGaussians   = numGaussians;
    	
		int inputDim    = windowSize * kmer + numGaussians;
		
		/* Input vector (only 1s for local coding) */
        inputVector = new double[maxSeqLen][inputDim];
        /* Index for local coding */
        inputIndex  = new short[maxSeqLen][inputDim];
        
        /* Number of inputs */
        this.numInputs = (int)Math.pow(nSymbols, kmer);
        
        this.sequenceCodingLength   =  numInputs * windowSize;
		
    }
    
    /**
     * Creates the input matrix from a LSTMSequence
     * 
     * @see LSTMSequence
     * 
     * @param lstmSequence
     * @param windowSize  
     * @param nSymbols
     */
    public void computeInputs(LSTMSequence lstmSequence) {
        
        /* half of the window */
        int wside = (windowSize - 1) / 2;
        
        int seqLength              = lstmSequence.getLocalCodings().length;
        
        int[] localCodings         = lstmSequence.getLocalCodings();
        double[][] positionCodings = lstmSequence.getPositionCodings();
        int[][] positionIndices    = lstmSequence.getPositionIndices();
        
        for (int i = 0; i < seqLength - kmer; i++) {
                 
            int starti = i - wside;

            /* amino coding */
            for (int j = 0; j < windowSize; j++) {
                int iter = starti + j;
                if (iter < 0 || iter >= seqLength) {
                    inputVector[i][j] = 0;
                    inputIndex[i][j] = 0;          
                }
                else {
                    int localCoding = localCodings[iter];
                    inputVector[i][j] = 1.0;
                    inputIndex[i][j] = (short)(j * numInputs + localCoding);
                }
            }
            
            for (int j = 0; j < numGaussians; j++) {
            	
            	inputVector[i][j + windowSize] = positionCodings[i][j];
            	inputIndex[i][j + windowSize]  = (short)(sequenceCodingLength + positionIndices[i][j]);
            	
            }
            
        }
            
    }
    
    public double[][] getInputVector() {
        return inputVector;
    }


    public short[][] getInputIndex() {
        return inputIndex;
    }

}
