/**
 * @author Martin Heusel
 * 
 * Copyright (C) 2008 Martin Heusel (mhe@bioinf.jku.at),
 *                      
 * Johannes Kepler University, Linz, Austria
 * Institute of Bioinformatics.
 * The software is maintained by Martin Heusel.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * If you use this library, please cite:
 *
 * @article{SeppHochreiter07152007,
 *   author = {Hochreiter, Sepp and Heusel, Martin and Obermayer, Klaus},
 *   title = {{Fast Model-based Protein Homology Detection without Alignment}},
 *   journal = {Bioinformatics},
 *   volume = {23},
 *   number = {14},
 *   pages = {1728-1736},
 *   doi = {doi:10.1093/bioinformatics/btm247},
 *   year = {2007},
 *   URL = {http://bioinformatics.oxfordjournals.org/cgi/content/abstract/btm247v1},
 *   eprint = {http://bioinformatics.oxfordjournals.org/cgi/reprint/btm247v1}
 * }
 * 
 * $Id: JLSTM.java 301 2009-04-21 12:20:09Z mhe $
 *
 */

package at.jku.bioinf.jlstmlogregression;

import java.util.ArrayList;
import java.util.Properties;

/**
 * Main class
 * 
 * @author mhe
 *
 */
public class JLSTM {
	
	// Property file
	static Properties properties;
	
	// Data containers for the sequences
	static ArrayList<LSTMSequence> seqsPosTrain;
	static ArrayList<LSTMSequence> seqsNegTrain;
	static ArrayList<LSTMSequence> seqsPosTest;
	static ArrayList<LSTMSequence> seqsNegTest;
	
	static ArrayList<LSTMSequence> seqsTrain;
	static ArrayList<LSTMSequence> seqsTest;
	
	static int windowSize;
	
	/* Number of symbols. 20 or 23 for Proteins and 4 for DNA. */
    static int nSymbols;
    
    /* kmer */
    static int kmer;
    
    /* for position coding */
    static int numPositionBins;
    static int numGaussians;
    
    /* Index of output unit */
    static int outputUnit;
    
    /* Number of LSTM blocks */
    static int numBlocks;
    
    /* Index of last input unit */
    static int lastInputUnit;
    
    /* Bias unit */
    static int biasUnit;
    
    /* Number of inputs for local coding */
    static int clampedInputs;
    
    /* Index of last LSTM unit */
    static int lastCellUnit;
    
    /* Number of units */
    static int numUnits;

    /* Blocksize for every block */
    static int[] blockSize;
    
    /* Range for random init of weights */
    static double initRange;
    
    /* Target values */
    static double targetPositive;
    static double targetNegative;
    
    /* Biases and initial weights */
    static double[] biasesInputgates;
    static double[] biasesOutputgates;
    static double[] biasesMemoryCellInputs;
    static double[] weightToOutput;
    static double outputBias;
    
    /* Learning rate */
    static double alpha;
    
    /* maximum number of epochs */
    static int maxEpochs;
    
    /* After testNEpochs a test is performed */
    static int testNEpochs;
    
    /* After writeWeightsAfterNEpochs the weight matrix is written */
    static int writeWeightsAfterNEpochs;
    
    /* Write an additional better readable weight matrix? For debugging
     * and investigation.
     */
    static int humanReadableWeightmatrix;
    
    /* Additional ROCN (e.g. N = 50) to be computed */
    static int rocn;

    /* Weight matrix */
    static WeightMatrix wm;
    
    static String weightFile;
    
    static String propertiesFile;
    
    /* test flag */
    static boolean test;
    
    static boolean loadWeights;
    
    /* Input matrix */
    static InputMatrix im;
    
    /* LSTM core */
    static LSTM lstm;
	
    /* Number of threads */
    static int numberThreads;
    
    static int maxSeqLen;
	
	/**
	 * Read and set properties from file
	 * 
	 */
	static void readProperties() {
	    
	    System.err.println();
        System.err.print("reading properties.. ");
		
	    // Read properties
	    properties = JLSTMProperties.readProperties(propertiesFile);

        windowSize    = Integer.parseInt(properties.getProperty("windowsize"));
        
        nSymbols      = Integer.parseInt(properties.getProperty("numberofsymbols"));
        
        kmer          = Integer.parseInt(properties.getProperty("kmer"));
        
        numPositionBins    = Integer.parseInt(properties.getProperty("numberofposbins"));
        numGaussians      = Integer.parseInt(properties.getProperty("numberofgaussians"));
        
        if (numPositionBins == 0 || numGaussians == 0) {
        	
        	numPositionBins = 0;
        	numGaussians    = 0;
        	
        }
        
        outputUnit       = Integer.parseInt(properties.getProperty("numberoutputunits"));
        numBlocks    = Integer.parseInt(properties.getProperty("numbermemorycellblocks"));
        initRange    = Double.parseDouble(properties.getProperty("halfintervallengthforinit"));
        
        targetPositive  = Double.parseDouble(properties.getProperty("targetvaluepos"));
        targetNegative  = Double.parseDouble(properties.getProperty("targetvalueneg"));
        
        alpha         = Double.parseDouble(properties.getProperty("learningrate"));
        
        maxEpochs     = Integer.parseInt(properties.getProperty("maxnumberofepochs"));
        
        testNEpochs   = Integer.parseInt(properties.getProperty("testafterNpochs"));
        
        writeWeightsAfterNEpochs  = Integer.parseInt(properties.getProperty("writeweightafterNepochs"));
        humanReadableWeightmatrix = Integer.parseInt(properties.getProperty("humanReadableWeightmatrix"));
                
        rocn          = Integer.parseInt(properties.getProperty("rocn"));
        
        // Blocksizes
        blockSize = new int[numBlocks];
        
        for (int i = 0; i < numBlocks; i++) {
            blockSize[i] = Integer.parseInt(properties.getProperty("blocksize" + String.valueOf(i + 1)));
        }
        
        // Input bias
        biasesMemoryCellInputs = new double[numBlocks];
        
        for (int i = 0; i < numBlocks; i++) {
            biasesMemoryCellInputs[i] = Double.parseDouble(properties.getProperty("initialinputbias" + String.valueOf(i + 1)));
        }

        // Inputgate biases
        biasesInputgates = new double[numBlocks];

        for (int i = 0; i < numBlocks; i++) {
            biasesInputgates[i] = Double.parseDouble(properties.getProperty("initialinputgatebias" + String.valueOf(i + 1)));
        }

        // Outputgate biases
        biasesOutputgates = new double[numBlocks];
        
        for (int i = 0; i < numBlocks; i++) {
            biasesOutputgates[i] = Double.parseDouble(properties.getProperty("initialoutputgatebias" + String.valueOf(i + 1)));
        }
        
        // Output weights
        weightToOutput = new double[numBlocks];
        
        for (int i = 0; i < numBlocks; i++) {
            weightToOutput[i] = Double.parseDouble(properties.getProperty("initialoutputweight" + String.valueOf(i + 1)));
        }

        // Output bias
        outputBias = Double.parseDouble(properties.getProperty("outputBias"));
        
        System.err.println("okay");
		
	}
	
	/**
	 * Read data files
	 * 
	 */
	static void readData() {
	    
	    System.err.println();
	    System.err.print("reading data.. ");
	    
	    ReadFasta rf = new ReadFasta(nSymbols, kmer, numPositionBins, numGaussians);
	    
	    // pos-train
	    seqsPosTrain = rf.read(properties.getProperty("inputdatatrainingpositive"), (byte)1);

	    maxSeqLen = rf.getMaxLenSeq();
	    
	    // neg-train	    
	    seqsNegTrain = rf.read(properties.getProperty("inputdatatrainingnegative"), (byte)0);

	    maxSeqLen = (maxSeqLen > rf.getMaxLenSeq()) ? maxSeqLen : rf.getMaxLenSeq();
	    
	    // pos-test
	    seqsPosTest = rf.read(properties.getProperty("inputdatatestpositive"), (byte)1);
	    
	    maxSeqLen = (maxSeqLen > rf.getMaxLenSeq()) ? maxSeqLen : rf.getMaxLenSeq();
	    
	    // neg-test
        seqsNegTest = rf.read(properties.getProperty("inputdatatestnegative"), (byte)0);

        maxSeqLen = (maxSeqLen > rf.getMaxLenSeq()) ? maxSeqLen : rf.getMaxLenSeq();
        
        System.err.print("balancing data.. ");
        // Read in again until data is balanced out
        while (seqsPosTrain.size() * 5 < seqsNegTrain.size()) {
            seqsPosTrain.addAll(rf.read(properties.getProperty("inputdatatrainingpositive"), (byte)1));
        }
        
        while (seqsNegTrain.size() * 5 < seqsPosTrain.size()) {
            seqsNegTrain.addAll(rf.read(properties.getProperty("inputdatatrainingnegative"), (byte)0));
        }
        
        System.err.println("okay");
        
        System.err.println();
        System.err.println("pos-train        :" + " " + seqsPosTrain.size());
        System.err.println("neg-train        :" + " " + seqsNegTrain.size());
        System.err.println("pos-test         :" + " " + seqsPosTest.size());
        System.err.println("neg-test         :" + " " + seqsNegTest.size());

        // Put negative and positive data together
        seqsTrain = new ArrayList<LSTMSequence>();
        
        seqsTrain.addAll(seqsPosTrain);
        seqsTrain.addAll(seqsNegTrain);
        
        seqsTest = new ArrayList<LSTMSequence>();
        
        seqsTest.addAll(seqsPosTest);
        seqsTest.addAll(seqsNegTest);
        
        System.err.println();
        System.err.println("train            :" + " " + seqsTrain.size());
        System.err.println("test             :" + " " + seqsTest.size());
        

        System.err.println();
        System.err.println("max. seq length  :" + " " + maxSeqLen);
		
	}
	
	/**
	 * Read only test data files
	 * 
	 */
	static void readDataTest() {
	    
	    System.err.println();
	    System.err.print("reading data for testing.. ");
	    
	    ReadFasta rf = new ReadFasta(nSymbols, kmer, numPositionBins, numGaussians);
		
	    // pos-test
	    seqsPosTest = rf.read(properties.getProperty("inputdatatestpositive"), (byte)1);
	    
	    maxSeqLen = rf.getMaxLenSeq();
	    
	    // neg-test
        seqsNegTest = rf.read(properties.getProperty("inputdatatestnegative"), (byte)0);
        
        // max. sequence length
        maxSeqLen = (maxSeqLen > rf.getMaxLenSeq()) ? maxSeqLen : rf.getMaxLenSeq();
        
        System.err.println("okay");
        
        System.err.println();
        System.err.println("pos-test         :" + " " + seqsPosTest.size());
        System.err.println("neg-test         :" + " " + seqsNegTest.size());

        // Put positive and negative data together
        
        seqsTest = new ArrayList<LSTMSequence>();
        
        seqsTest.addAll(seqsPosTest);
        seqsTest.addAll(seqsNegTest);
        
        System.err.println();
        System.err.println("test             :" + " " + seqsTest.size());
        
        System.err.println();
        System.err.println("max. seq length  :" + " " + maxSeqLen);
		
	}

	/**
	 * Init some stuff
	 * 
	 */
	static void init() {

        /* Number of inputs */
		int numInputs = (int)Math.pow(nSymbols, kmer) * windowSize + numPositionBins;
        
		lastInputUnit = numInputs - 1;  
        
        /* Bias unit */
        biasUnit = lastInputUnit + 1;
        
        /* Number of inputs for local coding */
        clampedInputs = windowSize;
        
        lastCellUnit = biasUnit;
        
        for (int i = 0; i < numBlocks; i++) {
            
          lastCellUnit += (2 + blockSize[i]);
        
        }
        
        lastCellUnit++;
        
        numUnits = lastCellUnit+outputUnit;
        
        System.err.println();
        int noi = lastInputUnit + 1;
        
        System.err.println("Number of Blocks : " + numBlocks);
        System.err.println("Window size      : " + windowSize);
        System.err.println();
        
        System.err.println("Position bins    : " + numPositionBins);
        System.err.println("Position gauss.  : " + numGaussians);
        System.err.println();
        
        System.err.println("kmers            : " + kmer);
        System.err.println();
        
        System.err.println("Number of Inputs : " + noi);
        //System.err.println("Bias Unit        : " + biasUnit);
        int nou = numUnits + 1;
        System.err.println("Number of Units  : " + nou);
        System.err.println();
        
        if (!test) {
        	System.err.println("Learning rate    : " + alpha);
        	System.err.println();
        	System.err.println("Stop after       : " + maxEpochs + " epochs");
        	System.err.println("Test every       : " + testNEpochs + " epochs");
        }
        
        System.err.println();
        System.err.println("ROCn             : " + rocn);
        
        System.err.println();
        System.err.println("Weight matrix    : " + weightFile);
        
        System.err.println();
        
        if (!test) {
        
        	if (humanReadableWeightmatrix != 0) {
        		System.err.println("Additional human readable weight matrix : yes");
        	}
        	System.err.println("Writing weight matrix every             : " + writeWeightsAfterNEpochs + " epochs");      
        
        	System.err.println();
        	System.err.println("Number of Threads                       : " + numberThreads);
        	
        }
        	
        // Create weightmatrix
        wm = new WeightMatrix(numUnits, lastCellUnit, lastInputUnit, biasUnit, numBlocks,
                              blockSize, numPositionBins, initRange);
        
        wm.setWeightfile(weightFile);
		
	}
	
	/**
	 * Parse command line options
	 * 
	 * @param args
	 */
	static void parseCommandline(String[] args) {

	    if (args.length < 6 || args.length > 7) {	        
	        usage();
	    }
	    
	    if (!"-p".equals(args[0])) {
	    	usage();
	    }
	    
	    else {
	    	propertiesFile = args[1];
	    }
	    
	    if (!"-w".equals(args[2])) {	
	    	usage();
	    }
	    
	    else {	
	    	weightFile = args[3];
	    }
	    
	    if (!"-t".equals(args[4])) {	
	    	usage();
	    }
	    
	    else {	
	    	numberThreads = Integer.valueOf(args[5]);
	    }
	     
	    
	    test        = false;
	    loadWeights = false;
	    
	    if (args.length == 7) {

        	if ("-test".equals(args[6])) {

        		test = true;

        	}
        	
        	else if ("-lw".equals(args[6])) {
        		
        		loadWeights = true;
        		
        	}
        	
        	else {
        		
        		usage();
        		
        	}
        	
	    }
		
	}
	
	
	static void usage() {
	    
	    System.err.println();
	    System.err.println("Usage: JLSTM -p properties file -w weightFile -t threads [-lw|-test]");
	    System.exit(-1);
	    
	}
	
	/**
	 * Main
	 * 
	 * @param args
	 */
	public static void main(String[] args) {
	    
	    parseCommandline(args);
			    
	    readProperties();
	    
	    if (test) {
	    	readDataTest();
	    }
	    else {
	    	readData();
	    }

        init();
        
        /* if -lw or -test parameter set then load weights */
        
        if (test || loadWeights) {
        	
        	System.err.println();

        	System.err.print("Reading weight matrix.. ");

        	wm.readWeightMatrix();

        	System.err.println("okay");
        
        	System.err.println();
        	
        }
        
        /* test with one thread */
        
        if (test) {
        	
        	LSTMThread lt = new LSTMThread(numUnits, lastInputUnit, biasUnit,
        			lastCellUnit, clampedInputs,
        			numBlocks, blockSize,
        			nSymbols, kmer, numGaussians, numPositionBins,
        			alpha,
        			biasesInputgates, biasesOutputgates, biasesMemoryCellInputs,
                    weightToOutput, outputBias,
                    windowSize,
        			targetPositive, targetNegative,
        			0, 0,
        			seqsPosTest.size(), seqsNegTest.size(),
        			maxSeqLen,
        			rocn,
        			maxEpochs,
        			testNEpochs,
        			writeWeightsAfterNEpochs,
        			null,
        			seqsTest,
        			wm, true);
        	
        	/* test and exit */
        	lt.start();
        	
        	try {
        		lt.join();
        	}
        	catch (InterruptedException ie) {
        	}
        	
        	System.exit(0);

	    }
        
        System.err.println();
        System.err.println("Running...");
        System.err.println();
        
        // Running the threads...
        
        ArrayList<Thread> threadList = new ArrayList<Thread>();
        
        for (int thread = 0; thread < numberThreads; thread++) {
        
        	LSTMThread lt = new LSTMThread(numUnits, lastInputUnit, biasUnit,
        			lastCellUnit, clampedInputs,
        			numBlocks, blockSize,
        			nSymbols, kmer, numGaussians, numPositionBins,
        			alpha,
        			biasesInputgates, biasesOutputgates, biasesMemoryCellInputs,
        			weightToOutput, outputBias,
        			windowSize,
        			targetPositive, targetNegative,
        			seqsPosTrain.size(), seqsNegTrain.size(),
        			seqsPosTest.size(), seqsNegTest.size(),
        			maxSeqLen,
        			rocn,
        			maxEpochs,
        			testNEpochs,
        			writeWeightsAfterNEpochs,
        			seqsTrain,
        			seqsTest,
        			wm, false);
        	
        	lt.start();
        	
        	threadList.add(lt);
        	
        	
        	/* Wait a little bit for the next thread */
            /*
        	try {
        		Thread.sleep(1000);
        	}
        	catch (InterruptedException ie){
        		// nothing 
        	}
            */
            
        }
        
        /* Wait until all threads have finished */
        for (int i = 0; i < threadList.size(); i++) {
        	
        	try {
        		threadList.get(i).join();
        	}
        	catch (InterruptedException ie) {
        		
        	}
        	
        }
        
        /* Start a last test */
        new LSTMThread(numUnits, lastInputUnit, biasUnit,
        		lastCellUnit, clampedInputs,
    			numBlocks, blockSize,
    			nSymbols, kmer, numGaussians, numPositionBins,
    			alpha,
    			biasesInputgates, biasesOutputgates, biasesMemoryCellInputs,
                weightToOutput, outputBias,
                windowSize,
    			targetPositive, targetNegative,
    			0, 0,
    			seqsPosTest.size(), seqsNegTest.size(),
    			maxSeqLen,
    			rocn,
    			maxEpochs,
    			testNEpochs,
    			writeWeightsAfterNEpochs,
    			seqsTrain,
    			seqsTest,
    			wm, true).start();
        
	}

}
