/**
 * @author Martin Heusel
 * 
 * Copyright (C) 2008 Martin Heusel (mhe@bioinf.jku.at),
 *                      
 * Johannes Kepler University, Linz, Austria
 * Institute of Bioinformatics.
 * The software is maintained by Martin Heusel.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * If you use this library, please cite:
 *
 * @article{SeppHochreiter07152007,
 *   author = {Hochreiter, Sepp and Heusel, Martin and Obermayer, Klaus},
 *   title = {{Fast Model-based Protein Homology Detection without Alignment}},
 *   journal = {Bioinformatics},
 *   volume = {23},
 *   number = {14},
 *   pages = {1728-1736},
 *   doi = {doi:10.1093/bioinformatics/btm247},
 *   year = {2007},
 *   URL = {http://bioinformatics.oxfordjournals.org/cgi/content/abstract/btm247v1},
 *   eprint = {http://bioinformatics.oxfordjournals.org/cgi/reprint/btm247v1}
 * }
 * 
 * $Id: LSTM.java 303 2009-04-22 08:20:03Z mhe $
 *
 */


package at.jku.bioinf.jlstmlogregression;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.text.DecimalFormat;
import java.util.ArrayList;

/**
 * This is the core of the LSTM. Computes forward, backward pass and errors.
 * 
 * @author mhe
 *
 */
public class LSTM {
    
    int numUnits;
    int lastInputUnit;
    int biasUnit;
    int lastCellUnit;
    int clampedInputs;
    int numBlocks;
    int[] blockSize;
    
    int nSymbols;
    
    int numGaussians;
    int numPositionBins;
    
    double[] ykOld;
    double[] ykNew;
    double[] yOutputgate;
    double[] yInputgate;
    double[][] yc;
    double[][] g;
    double[][] h;
    double[][] s;
    double[][][] sc;
    
    double[] e;
    double[] eo;
    double[][] ec;
    double[][] es;
    
    double[][][] si;
    
    double[] error;
    double targetPositive;
    double targetNegative;
    
    double sequenceError;
    boolean isSequenceCorrect;
    double epochError;
    
    int falsePositives;
    int falseNegatives;
    
    int numPositivesTraining;
    int numNegativesTraining;
    int numPositivesTest;
    int numNegativesTest;
    
    int numSequencesTraining;
    int numSequencesTest;
    
    boolean isPositive;
    
    double alpha;
    
    int rocn;
    
    ArrayList<RocData> rocList;
    
    ComputeRoc cr = new ComputeRoc();
    
    double[][] inputVector;
    short[][] inputIndex;

    WeightMatrix wm;
    
    double[][] weightMatrixLocal;
    double[][] weightMatrixGlobal;
    
    double[][] DW;
    
    public LSTM (int numUnits, int lastInputUnit, int biasUnit, 
    		     int lastCellUnit, int clampedInputs,
                 int numBlocks, int[] blockSize,
                 int nSymbols, int numgauss, int numposbins,
                 double alpha, 
                 double targetpositive, double targetnegative,
                 int numpostrain, int numnegtrain,
                 int numpostest, int numnegtest,
                 int rocn,
                 WeightMatrix wm) {
        
        this.numUnits             = numUnits;
        this.lastInputUnit        = lastInputUnit;
        this.biasUnit             = biasUnit;
        this.lastCellUnit         = lastCellUnit;
        this.clampedInputs        = clampedInputs;
        this.numBlocks            = numBlocks;
        this.blockSize            = blockSize;
    
        this.nSymbols             = nSymbols;
        
        this.numGaussians         = numgauss;
        this.numPositionBins      = numposbins;
        
        this.numPositivesTraining = numpostrain;
        this.numNegativesTraining = numnegtrain;
        this.numSequencesTraining = numpostrain + numnegtrain;

        this.numPositivesTest     = numpostest;
        this.numNegativesTest     = numnegtest;
        this.numSequencesTest     = numpostest + numnegtest;

        this.alpha                = alpha;
        
        this.targetPositive       = targetpositive;
        this.targetNegative       = targetnegative;
        
        this.rocn                 = rocn;
        
        this.wm                   = wm;
        
        this.weightMatrixGlobal   = wm.getWeightMatrix();
        
        // Init matrices
        
        DW    = new double[numUnits][numUnits];
        //DWold  = new double[numUnits][numUnits];
        
        ykOld = new double[numUnits + 1];
        ykNew = new double[numUnits + 1];
        
        yOutputgate  = new double[numBlocks];
        yInputgate   = new double[numBlocks];
        yc     = new double[numBlocks][5]; // TODO: change 5 to maximum blocksize
        g      = new double[numBlocks][5];
        h      = new double[numBlocks][5];
        s      = new double[numBlocks][5];
        
        si     = new double[numBlocks][5][numUnits + 1];
        sc     = new double[numBlocks][5][numUnits + 1];
        
        e      = new double[numUnits + 1];
        eo     = new double[numBlocks];
        ec     = new double[numBlocks][5];
        es     = new double[numBlocks][5];
        
        error  = new double[4];   // TODO: change 4 to number output units
        
        this.rocList = new ArrayList<RocData>(numSequencesTraining);
        
        this.sequenceError = 0;
        this.isSequenceCorrect = true;
        this.epochError = 0;
        
        this.falsePositives = 0;
        this.falseNegatives = 0;
        
        reset();
        
    }
    
    /**
     * Fast exponential approximation
     * 
     * {@link http://martin.ankerl.com/2007/02/11/optimized-exponential-functions-for-java/}
     * 
     * @param val
     * @return
     */
    public double fastexp(double val) {  
    	final long tmp = (long) (1512775 * val + (1072693248 - 60801));  
    	return Double.longBitsToDouble(tmp << 32);  
    }  
    
    /**
     * Compute the error
     */
    void compError () {
        
        /* compute error */
        for (int i = lastCellUnit; i < numUnits; i++) {
            
            error[0] = (this.isPositive) ? targetPositive - ykNew[i] : targetNegative - ykNew[i];
            
        }
        
        double err = 0;
        
        err +=  error[0] * error[0];
        
        sequenceError += err;
        epochError += err;
        
        if ((isSequenceCorrect) && (Math.abs(error[0]) > Math.abs(targetPositive - targetNegative)/2)) { 
        
          isSequenceCorrect = false;
          
          if (this.isPositive) { 
              falseNegatives++;
          }
          else {
              falsePositives++;
          }

        }
        
               
    }
    
    /**
     * Output of net performance (MSE, false-positives and false-negatives)
     * 
     * @param isTrain
     */
    public void printError(int epoch, boolean isTrain) {

        int numpos, numneg, numseqs;
        String what = "";
        String pre  = "";
        
        if (isTrain) {
        	numpos = numPositivesTraining;
        	numneg = numNegativesTraining;
        	numseqs = numSequencesTraining;
        	what    = "training";
        	pre     = "";
        	
        	cr.computeAUC(rocList, numPositivesTraining, numNegativesTraining, rocn);
        	
        }
        else {
        	numpos = numPositivesTest;
        	numneg = numNegativesTest;
        	numseqs = numSequencesTest;
        	what    = "test";
        	pre     = "TEST: ";
        	
        	cr.computeAUC(rocList, numPositivesTest, numNegativesTest, rocn);
        	
        }
        
        double auc = cr.getAuc();
        double aucn = cr.getAucn();
        
        double fnp = (numpos != 0) ? (double) falseNegatives / (double) numpos : 0;
        double fpp = (numneg != 0) ? (double) falsePositives / (double) numneg : 0;
        
        double mse = epochError / (1.0 * numseqs);
        
        BufferedWriter out = null;

        try {
          
          out = new BufferedWriter( new FileWriter( "out.txt", true ) );
          
          out.write("epoch: " + epoch);
          out.newLine();
          out.write(pre + "MSE ");
          out.write(new DecimalFormat("#.#####").format(mse));
          out.newLine();
          out.write(pre + "false-negative:");
          out.write(falseNegatives + " (out of " + numpos + " positive " + what + " examples) ");
          out.write(new DecimalFormat("#0.0").format(fnp * 100) + "%");
          out.newLine();
          out.write(pre + "false-positive:");
          out.write(falsePositives + " (out of " + numneg + " negative " + what + " examples) ");
          out.write(new DecimalFormat("#0.0").format(fpp * 100) + "%");
          out.newLine();
          out.write(pre +  "ROC ");
          out.write(new DecimalFormat("#0.000").format(auc));
          out.newLine();
          out.write(pre +  "ROC" + rocn + " ");
          out.write(new DecimalFormat("#0.000").format(aucn));
          out.newLine();
          out.newLine();
          out.close();
          
        }
        catch ( IOException e ) {
          System.err.println( e );
        }
        finally {
          try {
            if ( out != null ) out.close();
          } catch ( IOException e ) { }
        }
        
        System.out.println("epoch: " + epoch);
        System.out.printf("%sMSE:%f\n", pre, mse);
        System.out.printf("%sfalse-negative:%d (out of %d positive %s examples) %.1f%%\n", pre, falseNegatives, numpos, what, fnp * 100);
        System.out.printf("%sfalse-positive:%d (out of %d negative %s examples) %.1f%%\n", pre, falsePositives, numneg, what, fpp * 100);
        System.out.printf("%sROC %.3f\n", pre, auc);
        System.out.printf("%sROC" + rocn + " %.3f\n", pre, aucn);
        System.out.println();
    	
    }
    
    /**
     * Timestep
     * 
     */
    public void timeforward() {
        
        for (int i = 0; i < numUnits; i++) {
            ykOld[i] = ykNew[i];
        }
        
        sequenceError = 0;
        isSequenceCorrect = true;
        
    }
    
    public void cloneWeightMatrix() {
    	
    	//this.weightMatrixLocal = this.weightMatrixGlobal.clone();
    	this.weightMatrixLocal = this.weightMatrixGlobal;
    	
    }
    
    /**
     * Forward pass
     * 
     * @param element
     * @param targ
     */
    public void forwardPass(int element, boolean targ) {
    	
        /* ### memory cells ### */

        int i = biasUnit + 1;
        
        for (int u = 0; u < numBlocks; u++) {

            double sum;
            
            yInputgate[u] = 1.0;
            ykNew[i]= yInputgate[u];

            /* output gates */
            
            i++;
            
            yOutputgate[u]  = 1.0;
            
            ykNew[i] = yOutputgate[u];

            /* uth memory cell block */
            for (int v = 0; v < blockSize[u]; v++) {
                
                /* activation of function g of vth memory cell of block u  */
                i++;
                sum = weightMatrixLocal[i][biasUnit]; // bias

                for (int j = 0; j < clampedInputs; j++) {
                    sum += weightMatrixLocal[i][inputIndex[element][j]] * inputVector[element][j];
                }
                
                /* position infos (if any) */
                for (int j = 0; j < numGaussians; j++) {
                	sum += weightMatrixLocal[i][inputIndex[element][clampedInputs + j]] * inputVector[element][clampedInputs + j];
                }

                g[u][v] = sum;
                
                /* update internal state  */
                s[u][v] += g[u][v];
                
                /* activation function h */
                h[u][v] = s[u][v];

                /* activation of vth memory cell of block u */
                yc[u][v] = h[u][v];
                ykNew[i] = yc[u][v];
                
                //System.out.println(i +  " " + ykNew[i]);
                
                
            }
            
            i++;
            
        }



        /* ### output units activation ### */

        if (targ) /* only if target for this input */ {

            for (int k = lastCellUnit; k < numUnits; k++) {        
                
                /* bias */
                double sum = weightMatrixLocal[k][biasUnit];
                
                /* memory cells input */
                i = biasUnit + 1;
                for (int u = 0; u < numBlocks; u++) {
                    i++;
                    for (int v = 0; v < blockSize[u]; v++) {
                        i++;
                        sum += weightMatrixLocal[k][i] * ykNew[i];
                        //System.out.println(i +  " " + weightMatrixLocal[k][i] + " " + ykNew[i]);
                    }
                    i++;
                }
                
                /* activation */
                ykNew[k] = 1/(1+fastexp(-sum));
                //System.out.println(ykNew[k] + " " + label);
                
                
                /* Store output and label in RocData container */
                byte label = (byte) ((isPositive) ? 1 : 0);
                RocData rd = new RocData(ykNew[k], label);
                
                rocList.add(rd);
                

            }       

        }

        
    }
    
    
    /**
     * Backward pass
     * 
     * @param element
     */
    public void backwardPass(int element, ArrayList<Integer> usedInputIndices) {
    	        
        /* output units */
        for (int k = lastCellUnit, j = 0; k < numUnits; k++, j++) {
            
            e[k] = error[j];
            
            /* weight update contribution */
            
            double tmp = alpha * e[k];
            
            /* bias to output unit */
            DW[k][biasUnit] += tmp;
            
        }
        
        /* error to memory cells ec[][] and internal states es[][] */

        int i = biasUnit + 1;
        
        /* input gates */
        for (int u = 0; u < numBlocks; u++) {
            
            i++;
            
            for (int v = 0; v < blockSize[u]; v++) {
                
                i++;
                
                double sum = 0;
                
                for (int k = lastCellUnit; k < numUnits; k++) {
                    
                    sum += weightMatrixLocal[k][i] * e[k];
                    
                }
                
                ec[u][v] = sum;
                
                es[u][v] = sum;
                
            }
            
            i++;
            
        }

        /* output gates */
        

        /* Derivatives of the internal state */

        derivatives(element);


        /* updates for weights to input and output gates and memory cells */
        
        i = biasUnit + 1;
        
        for (int u = 0; u < numBlocks; u++) {        
            
            /* input gates */
            
        	
            /*output gates */
            
            i++;
            
            
            /* memory states */
            for (int v = 0; v < blockSize[u]; v++) {
                
                i++;
                
                double tmp = alpha * es[u][v];
                
                /* input from k-mers found in sequence */
                for (int j = 0; j < usedInputIndices.size(); j++) {
                	
                	int uii = usedInputIndices.get(j);
                	
                    DW[i][uii] += tmp*sc[u][v][uii];
                    
                    //System.out.println(j + " " + uii + " " + tmp + " " + sc[u][v][uii]);
                    
                }
                
                for (int j = 0; j < numPositionBins; j++) {
                	
                	DW[i][biasUnit - numPositionBins + j] += tmp * sc[u][v][biasUnit - numPositionBins + j]; 
                	
                }
                
                // bias
                DW[i][biasUnit] += tmp*sc[u][v][biasUnit];
                
            }
            
            i++;
            
        }

        
    }
    
    /**
     * Computes derivatives
     * 
     * @param element
     */
    public void derivatives(int element) {
              
        for (int u = 0; u < numBlocks; u++) {
            
            //tmp0 = (1.0-Y_in[u])*Y_in[u];
            //tmp1 = Y_in[u]*2.0;
            
            for (int v = 0; v < blockSize[u]; v++) {
                
                //double tmp1 = g[u][v];
                
                //double tmp2 = 0;
                
                //double tmp3 = (1.0 - 0.5 * g[u][v]) * tmp1;
                
                for (int j = 0; j < clampedInputs; j++) {
                    
                    /* weights to input gate */
                    //si[u][v][inputIndex[element][j]] += tmp2 * inputVector[element][j];
                    
                    /* weights to cell input */
                    //sc[u][v][inputIndex[element][j]] += ykOld[inputIndex[element][j]];
                    sc[u][v][inputIndex[element][j]] += inputVector[element][j];
                    //System.out.println(element + " " + j + " " + tmp3 + " " + inputVector[element][j]);
                
                }
                
                
                /* position inputs */
                for (int j = 0; j < numGaussians; j++) {
                	
                	sc[u][v][inputIndex[element][clampedInputs + j]] += inputVector[element][clampedInputs + j];
                	
                }
                
                /* bias to memory cell */
                sc[u][v][biasUnit] += 1.0;
                
                
            }
            
        }
        
    }
    
      
    /**
     * Initialization of the net
     * 
     * @param biasesInputgates
     * @param biasesOutputgates
     * @param biasesMemoryCellInputs
     * @param weightToOutput
     * @param outputBias
     */
    public void init (double[] bias_inp, double[] bias_out,
                      double[] bias_memin, double[] weight_memout,
                      double outputbias) {
        
    	cloneWeightMatrix();
    	
        /* bias and weight to output unit initalizations */
        int i = biasUnit + 1;
        for (int u = 0; u < numBlocks; u++) {
            /* bias to input gate */
            weightMatrixLocal[i][biasUnit] = bias_inp[u];
            i++;
            /* bias to output gate */
            weightMatrixLocal[i][biasUnit] = bias_out[u];
            for (int v = 0; v < blockSize[u]; v++) {
                i++;
            }
            /* bias to memory cell */
            weightMatrixLocal[i][biasUnit] = bias_memin[u];
            /* weight to output unit */ 
            weightMatrixLocal[lastCellUnit][i] = weight_memout[u];

            i++;
        }

        /* output bias */
        weightMatrixLocal[lastCellUnit][biasUnit] = outputbias;

        
    }
    
    /**
     * Reset the net
     * 
     */
    public void reset () {
        
        for (int i = 0; i < numUnits; i++) {
            
            ykNew[i] = 0.5;
            ykOld[i] = 0.5;
            
            for (int j = 0; j < numUnits; j++) {
            	DW[i][j] = 0;
            }
            
        }
        
        ykNew[biasUnit] = 0.0;
        ykOld[biasUnit] = 0.0;
        
        for (int u = 0; u < numBlocks; u++) {
            
            yInputgate[u]  = 0;
            yOutputgate[u] = 0;
            
            for (int v = 0; v < blockSize[u]; v++) {
                
                s[u][v]  = 0;
                g[u][v]  = 0;
                h[u][v]  = 0;
                yc[u][v] = 0.5;
                
                for (int j = 0; j < lastCellUnit; j++) {
                    
                    si[u][v][j] = 0;
                    sc[u][v][j] = 0;
                    
                }
                
           }
            
        }

    }
    
    public void setInputVector(double[][] inputVector) {
        this.inputVector = inputVector;
    }

    public void setInputIndex(short[][] inputIndex) {
        this.inputIndex = inputIndex;
    }


    public void setIsPositive (boolean ispositive) {
        this.isPositive = ispositive;
    }

	public void setEpoch_err(double epoch_err) {
		this.epochError = epoch_err;
	}

	public void setFalsepos(int falsepos) {
		this.falsePositives = falsepos;
	}

	public void setFalseneg(int falseneg) {
		this.falseNegatives = falseneg;
	}
    
	public double[][] getDW() {
		return DW;
	}
	
}
