/**
 * @author Martin Heusel
 * 
 * Copyright (C) 2008 Martin Heusel (mhe@bioinf.jku.at),
 *                      
 * Johannes Kepler University, Linz, Austria
 * Institute of Bioinformatics.
 * The software is maintained by Martin Heusel.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * If you use this library, please cite:
 *
 * @article{SeppHochreiter07152007,
 *   author = {Hochreiter, Sepp and Heusel, Martin and Obermayer, Klaus},
 *   title = {{Fast Model-based Protein Homology Detection without Alignment}},
 *   journal = {Bioinformatics},
 *   volume = {23},
 *   number = {14},
 *   pages = {1728-1736},
 *   doi = {doi:10.1093/bioinformatics/btm247},
 *   year = {2007},
 *   URL = {http://bioinformatics.oxfordjournals.org/cgi/content/abstract/btm247v1},
 *   eprint = {http://bioinformatics.oxfordjournals.org/cgi/reprint/btm247v1}
 * }
 * 
 * $Id: LSTMThread.java 301 2009-04-21 12:20:09Z mhe $
 *
 */


package at.jku.bioinf.jlstmlogregression;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Random;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;


/**
 * LSTM Thread for running a thread
 * Has it's own lstm core and sequence data containers
 * Weight matrix is shared across all threads only the weight update
 * is synchronized.
 * 
 * @author mhe
 *
 */
public class LSTMThread extends Thread {
	
	ArrayList<LSTMSequence> seqsTrain;
	ArrayList<LSTMSequence> seqsTest;
    
	/* Input matrix */
    InputMatrix im;
    
    /* LSTM core */
    LSTM lstm;
    
    /* Weight matrix */
    WeightMatrix wm;
    
    int kmer; 
    
    /* Maximum number of epochs before stop */
    int maxEpochs;
    
    /* After testNEpochs a test is performed */
    int testnepochs;
    
    /* After writeWeightsAfterNEpochs the weight matrix is written */
    int writeweightsafternepochs;
    
    /* Write an additional better readable weight matrix? For debugging
     * and investigation.
     */
    int humanreadableweightmatrix;
    
    boolean test;
    
    /* Additional ROCN (e.g. N = 50) to be computed */
    int rocn;
    
    /* epoch is shared across all threads */
    static int epoch;
    
    ReentrantReadWriteLock rwl = new ReentrantReadWriteLock();
    private final Lock wlock = rwl.writeLock();
    
    public LSTMThread(int numUnits, int lastInputUnit, int biasUnit,
    		int lastCellUnit, int clampedInputs,
            int numBlocks, int[] blockSize,
            int nSymbols, int kmer, int numGaussians, int numPositionBins,
            double alpha,
            double[] biasesInputgates, double[] biasesOutputgates, double[] biasesMemoryInputs,
            double[] weightsToOutputs, double outputBias,
            int windowSize,
            double targetPositive, double targetNegative,
            int numPositiveTraining, int numNegativeTraining,
            int numPositiveTest, int numNegativeTest,
            int maxSeqLen,
            int rocn,
            int maxEpochs,
            int testNEpochs,
            int writeWeightsAfterNEpochs,
            ArrayList<LSTMSequence> seqsTrain,
            ArrayList<LSTMSequence> seqsTest,
            WeightMatrix wm, boolean test) {
    	
        
        this.kmer           = kmer;
        
        this.maxEpochs      = maxEpochs;
        
        this.testnepochs    = testNEpochs;
        
        this.writeweightsafternepochs = writeWeightsAfterNEpochs;
        
        this.test           = test;
        
        if (!test) {     
        	this.seqsTrain      = (ArrayList<LSTMSequence>)seqsTrain.clone();
        }
        
        this.seqsTest       = (ArrayList<LSTMSequence>)seqsTest.clone();
        
        this.wm             = wm;
        
    	// Create inputmatrix
        im = new InputMatrix(windowSize, nSymbols, kmer, numGaussians, maxSeqLen);
        
        // Make LSTM network
        if (test) {
        	      	
        	lstm = new LSTM(numUnits, lastInputUnit, biasUnit,
        			 lastCellUnit, clampedInputs,
                     numBlocks, blockSize,
                     nSymbols, numGaussians, numPositionBins,
                     alpha,
                     targetPositive, targetNegative,
                     0, 0,
                     numPositiveTest, numNegativeTest,
                     rocn,
                     wm);
        	
        }
        else {
        	
        	lstm = new LSTM(numUnits, lastInputUnit, biasUnit,
        			lastCellUnit, clampedInputs,
        			numBlocks, blockSize,
        			nSymbols, numGaussians, numPositionBins,
        			alpha,
        			targetPositive, targetNegative,
        			numPositiveTraining, numNegativeTraining,
        			numPositiveTest, numNegativeTest,
        			rocn,
        			wm);
        	
        	lstm.init (biasesInputgates, biasesOutputgates,biasesMemoryInputs, weightsToOutputs, outputBias);
        	
        }
        	
        
    	
    }
	
	/**
	 * Training
	 * 
	 * @param epoch
	 */
	void training(int epoch) {
		
		/* Compute for every sequence the indices for k-mers found in the sequence
		   Usually that is much smaller than nSymbols ^ k. So it's a speed-up for
		   backwardpass and weight update.
		   This are the datacontainers for removing not used indices for every
		   sequence.
		*/	
		HashMap<Integer, Integer> usedInputIndicesHM = new HashMap<Integer, Integer>();
		ArrayList<Integer> usedInputIndices = new ArrayList<Integer>();
		
		// For all training sequences
        for (int i = 0; i < seqsTrain.size(); i++) {
        	
            // Create input matrix
            im.computeInputs(seqsTrain.get(i));
            
            usedInputIndicesHM.clear();
            usedInputIndices.clear();
            
            // Collect all input indices (k-mers) in current sequence
            // in a HashMap for removing redundant indices
            for (int j = 0; j < seqsTrain.get(i).getLocalCodings().length - 1 - kmer; j++) {
            
            	int index = seqsTrain.get(i).getLocalCodings()[j];
            	usedInputIndicesHM.put(index, index);
            	
            }
            
            // Put unique input indices in an ArrayList (remove redundant indices)
            for (int j : usedInputIndicesHM.keySet()) {
            	
            	usedInputIndices.add(usedInputIndicesHM.get(j));
            	
            }
            
            Collections.sort(usedInputIndices);
            
            // Set  input index and input vector in the lstm core
            lstm.setInputIndex(im.getInputIndex());
            lstm.setInputVector(im.getInputVector());
            
            // Set label 
            lstm.setIsPositive (seqsTrain.get(i).label == 1);
            
            // For all elements in the current sequence i except the last one
            int j;
            for (j = 0; j < seqsTrain.get(i).getLocalCodings().length - 1 - kmer; j++) {
            	lstm.cloneWeightMatrix();
                lstm.forwardPass(j, false);
                lstm.derivatives(j);
                lstm.timeforward();
            }
            
            
            
            // The last element has a target
            lstm.cloneWeightMatrix();
            lstm.forwardPass(j, true);
            lstm.compError();
            lstm.backwardPass(j, usedInputIndices);
            
            
            
            
            /* write lock the weight update */ 
            wlock.lock();
            wm.weightUpdate(lstm.getDW(), usedInputIndices);
            wlock.unlock();
            
            lstm.reset();          
        
        }
        
        lstm.printError(epoch, true);
        lstm.setEpoch_err(0);
        lstm.setFalsepos(0);
        lstm.setFalseneg(0);		
		
		
	}
	
	/**
	 * Test
	 * 
	 * @param epoch
	 */
	void test(int epoch) {
		
    	// For all test sequences
    	for (int i = 0; i < seqsTest.size(); i++) {

    		im.computeInputs(seqsTest.get(i));
    		
    		lstm.setInputIndex(im.getInputIndex());
            lstm.setInputVector(im.getInputVector());
            
    		
    		lstm.setIsPositive(seqsTest.get(i).label == 1);
    		
    		lstm.cloneWeightMatrix();
    		
    		int j;
    		
    		for (j = 0; j < seqsTest.get(i).getLocalCodings().length - 1 - kmer; j++) {
    			lstm.forwardPass(j, false);
    			lstm.timeforward();
    		}
    		
    		lstm.forwardPass(j, true);
    		lstm.compError();
    		lstm.reset();           

    	}
    	
    	lstm.printError(epoch, false);
    	lstm.setEpoch_err(0);
    	lstm.setFalsepos(0);
    	lstm.setFalseneg(0);
		
		
	}	
	
	
	/* (non-Javadoc)
	 * @see java.lang.Thread#run()
	 */
	@Override public void run() {
		
		int myepoch;
		
		if (test) {
			
			test(epoch);
			
			return;
			
		}
		
		Lock lock = new ReentrantLock();
		
		while (epoch <= maxEpochs - 1) {
			
			lock.lock();
            
            epoch++;
            myepoch = epoch;
            
            lock.unlock();
            
            // Shuffle data set
            Collections.shuffle(seqsTrain, new Random());
            
            // Training
            
            training(myepoch);
            
            // Test after n epochs
            
            if ((myepoch % testnepochs == 0 && myepoch > 0) || myepoch == maxEpochs) {
            
            	test(myepoch);
            
            }
            
            // Write weight matrix
            
            if ((myepoch % writeweightsafternepochs == 0 && myepoch > 0) || myepoch == maxEpochs) {
                
                wm.writeWeightMatrix();
                
                if (humanreadableweightmatrix != 0) {
                    wm.writeReadableWeightMatrix();
                }
                
            }

        }
		
	}
	

}
