/**
 * 
 * Copyright (C) 2008 Martin Heusel (mhe@bioinf.jku.at),
 *                      
 * Johannes Kepler University, Linz, Austria
 * Institute of Bioinformatics.
 * The software is maintained by Martin Heusel.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * If you use this library, please cite:
 *
 * @article{SeppHochreiter07152007,
 *   author = {Hochreiter, Sepp and Heusel, Martin and Obermayer, Klaus},
 *	 title = {{Fast Model-based Protein Homology Detection without Alignment}},
 *	 journal = {Bioinformatics},
 *	 volume = {23},
 *	 number = {14},
 *	 pages = {1728-1736},
 *	 doi = {doi:10.1093/bioinformatics/btm247},
 *	 year = {2007},
 *	 URL = {http://bioinformatics.oxfordjournals.org/cgi/content/abstract/btm247v1},
 *	 eprint = {http://bioinformatics.oxfordjournals.org/cgi/reprint/btm247v1}
 * }
 * 
 * $Id: ReadFasta.java 307 2009-04-28 05:53:45Z mhe $
 *
 */

package at.jku.bioinf.jlstmlogregression;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.NoSuchElementException;

import org.biojava.bio.BioException;
import org.biojavax.bio.seq.RichSequence;
import org.biojavax.bio.seq.RichSequenceIterator;
import org.biojavax.bio.seq.RichSequence.IOTools;

/**
 * Reads a FASTA file and computes indices for sparse local coding.
 * 
 * @author mhe
 */
public class ReadFasta {
	
	int kmer;
	int numPositionBins;
	int numGaussians;
	
	int maxLenSeq;
	
	/*
	 * DNA alphabet
	 */
	final char[] alphabet = {'A', 'T', 'C', 'G'};
    
	/*
	 * Protein alphabet complete
	 */
	/*
	final char alphabet[] =
    {'A', 'R', 'N', 'D', 'C', 'Q', 'E', 'G', 'H', 'I', 'L', 'K', 'M',
     'F', 'P', 'S', 'T', 'W', 'Y', 'V', 'B', 'Z', 'X'}; 
	*/
	
	/**
	 * Protein alphabet without B and Z
	 */
	/*
	final char alphabet[] =
    {'A', 'R', 'N', 'D', 'C', 'Q', 'E', 'G', 'H', 'I', 'L', 'K', 'M',
     'F', 'P', 'S', 'T', 'W', 'Y', 'V', 'X'}; 
	*/
	
    HashMap<String, Integer> hmIndices;
    
    ArrayList<GaussStruct> positionVector;
    
    double[][] positionCodings;
    int[][] positionIndices;
    
    public ReadFasta(int nSymbols, int kmer, int numposbins, int numgauss) {
    	
		this.kmer       = kmer;
		this.numPositionBins = numposbins;
		this.numGaussians   = numgauss;
		
		this.positionVector = new ArrayList<GaussStruct>(numposbins);
		
		for (int i = 0; i < numposbins; i++) {
			
			GaussStruct gs = new GaussStruct();
			gs.gauss = 0;
			gs.idx   = 0;
			positionVector.add(gs);
			
		}
		
    	computeIndices(nSymbols, kmer);
    	
    }
    
    /*
     * Data structure that holds a gauss value and an
     * index for an input vector.
     */
    class GaussStruct implements Comparable<Object> {
    	
    	double gauss;
    	int idx;
    	
    	public int compareTo(Object obj) {
    		
    	    GaussStruct gs = (GaussStruct) obj;
    	    
    	    if ( gauss == gs.gauss ) {
    			return 0;
    		}
    		else if ( gauss > gs.gauss) {
    			return -1;
    		}
    		else return 1;
    	    
    	}
    	
    }
    
    public class GaussStructComparator implements Comparator<Object> {
    	  
    	public int compare(Object obj1, Object obj2) {
    	
    		GaussStruct gs1 = (GaussStruct) obj1;
    	    GaussStruct gs2 = (GaussStruct) obj2;

    	    int gsComp = gs1.compareTo(gs2);

    	    return ((gsComp == 0) ? gs1.compareTo(gs2) : gsComp);
    	  
    	  }
    	  
    }
	

    /**
     * Computes indices for all k-mers and puts them in a HashMap.
     * 
     * @param nSymbols
     * @param k
     */
    private void computeIndices(int nSymbols, int k) {
    	
    	this.hmIndices = new HashMap<String, Integer>();
    	
    	char[] kmer = new char[k];
    	
    	for (int i = 0; i < Math.pow(nSymbols, k); i++) {
    		for (int n = 0; n < k; n++) {
    			int pow = 1;
    			for (int m = 0; m < n; m++) {
    				pow *= nSymbols;
    			}
    			
    			/* AAAA TAAA CAAA GAAA ATAA TTAA ... */
    			kmer[n] = alphabet[ i / pow % nSymbols];
    		}
    		String kmerString = String.copyValueOf(kmer);
    		hmIndices.put(kmerString, i);
    		//System.out.println(i + " " + kmerString);
    			
    	}
    	
    }
    
	double lstmgauss(double m, double s, double x) {
		   return (4.0 / s) * Math.exp(-0.5 * Math.pow((x - m) / (0.5 * s),2));
	}

    
    private void computePositionCodings( RichSequence rs ) {
    	
    	positionCodings = new double[rs.length()][numGaussians];
    	positionIndices = new int[rs.length()][numGaussians];
    	
    	if (numPositionBins > 0) {

    		double sd = (double)rs.length() / (numPositionBins - 1.0);

    		for (int i = 0; i < rs.length(); i++) {
    			for (int j = 0; j < numPositionBins; j++) {
    				this.positionVector.get(j).gauss = lstmgauss((double)j * sd, sd, (double)i);
    				this.positionVector.get(j).idx = j;
    			}
    			Collections.sort(this.positionVector, new GaussStructComparator());
    			positionCodings[i][numGaussians / 2] = positionVector.get(0).gauss;
    			positionIndices[i][numGaussians / 2] = positionVector.get(0).idx;
    			for (int j = 1, k = 1; j < numGaussians / 2 + 1; j++, k += 2) {
    				//System.out.println(j + " " + k + " " + numGaussians / 2);
    				positionCodings[i][numGaussians / 2 - j] = positionVector.get(k).gauss;
    				positionIndices[i][numGaussians / 2 - j] = positionVector.get(k).idx;
    				positionCodings[i][numGaussians / 2 + j] = positionVector.get(k + 1).gauss;
    				positionIndices[i][numGaussians / 2 + j] = positionVector.get(k + 1).idx;
    				//System.out.printf("%d %f %d %f %d %f %d\n", i, positionCodings[i][numGaussians / 2 - j], positionIndices[i][numGaussians / 2 - j], positionCodings[i][numGaussians / 2],  positionIndices[i][numGaussians / 2], positionCodings[i][numGaussians / 2 + j], positionIndices[i][numGaussians / 2 + j]);  
    			}
    		}

    	}
    	    	
    }

    
 
    /**
     * Reads a FASTA file with help of Biojava RichSequence.IOTools.
     * Creates a LSTMSequence for any sequence
     * and label and stores it in an ArrayList which is returned. 
     * 
     * @param filename
     * @param label
     * @return ArrayList<LSTMSequence>
     */
    public ArrayList<LSTMSequence> read (String filename, byte label) {
        
        ArrayList<LSTMSequence> seqs = new ArrayList<LSTMSequence>();
        
        char[] kmerChars = new char[kmer];
        
        try {
            // Setup file input
            BufferedReader br =
                new BufferedReader(new FileReader(filename));
            
            // RichSequenceIterator
            RichSequenceIterator rsi = IOTools.readFastaProtein(br, null);
            
            // For each sequence do
            while(rsi.hasNext()) {
            	
                try {
                    
                    RichSequence rs = rsi.nextRichSequence();
                    
                    //System.out.println(rs.getName() + " " + rs.getDescription());
                    
                    // Sequence data
                    char[] seq = new char[rs.length()];
                    int[] localCodings = new int[rs.length()];
                    
                    rs.seqString().getChars(0, rs.length(), seq, 0);
                    
                    // Compute local coding indices
                    
                    for (int i = 0; i < rs.length() - kmer; i++) {
                    	for (int j = 0; j < kmer; j++) {
                    		kmerChars[j] = seq[i + j];
                    		//System.out.print(kmerChars[j]);
                    	}
                    	//System.out.println();
                    	String kmerString = String.copyValueOf(kmerChars);
                    	//System.out.println(hmIndices.get(kmerChars));
                        localCodings[i] = hmIndices.get(kmerString);
                    }
                    
                    computePositionCodings(rs);
                    
                    // Store codings and label in a LSTMSequence
                    LSTMSequence lstmseq = new LSTMSequence(localCodings, positionCodings.clone(), positionIndices.clone(), label);
                    
                    // Add LSTMSequence to ArrayList
                    seqs.add(lstmseq);
                    
                    this.maxLenSeq = (rs.length() > maxLenSeq) ? rs.length() : maxLenSeq;
                    
                }
                catch (BioException ex) {
                    //not in fasta format or wrong alphabet
                    System.out.println("not in fasta format or wrong alphabet");
                    System.exit(0);
                }
                catch (NoSuchElementException ex) {
                    //no fasta sequences in the file
                    System.out.println("no fasta sequences in the file");
                    System.exit(0);
                }
                
            }
            
            br.close();
            
        }
        catch (FileNotFoundException fnfex) {
            //problem reading file
            System.err.println("jLSTM: Could not open sequence file: " + filename);
            System.exit(1);
        }
        
        catch (IOException ioe) {
        	  //problem reading file
            System.err.println("jLSTM: Could not close sequence file: " + filename);
            //System.exit(1);
        }
        
        return seqs;
        
    }
    
    public int getMaxLenSeq() {
		return maxLenSeq;
	}

    
}
