/**
 * @author Martin Heusel
 * 
 * Copyright (C) 2008 Martin Heusel (mhe@bioinf.jku.at),
 *                      
 * Johannes Kepler University, Linz, Austria
 * Institute of Bioinformatics.
 * The software is maintained by Martin Heusel.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * If you use this library, please cite:
 *
 * @article{SeppHochreiter07152007,
 *   author = {Hochreiter, Sepp and Heusel, Martin and Obermayer, Klaus},
 *   title = {{Fast Model-based Protein Homology Detection without Alignment}},
 *   journal = {Bioinformatics},
 *   volume = {23},
 *   number = {14},
 *   pages = {1728-1736},
 *   doi = {doi:10.1093/bioinformatics/btm247},
 *   year = {2007},
 *   URL = {http://bioinformatics.oxfordjournals.org/cgi/content/abstract/btm247v1},
 *   eprint = {http://bioinformatics.oxfordjournals.org/cgi/reprint/btm247v1}
 * }
 * 
 * $Id: WeightMatrix.java 308 2009-04-29 07:11:59Z mhe $
 *
 */

package at.jku.bioinf.jlstmlogregression;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.InputMismatchException;
import java.util.Random;
import java.util.Scanner;

/**
 * Represents a weight matrix with some methods.
 * 
 * @author mhe
 *
 */
public class WeightMatrix {
    
    double[][] weightMatrix;
    
    int numUnits;
    int lastInputUnit;
    int biasUnit;
    int lastCellUnit;
    int numBlocks;
    int numPositionBins;
    
    int[] blockSize;
    
    
    int wmdimx;
    int wmdimy;
    int wmdimyoffset;
    
    String weightFile;
    
    public WeightMatrix (int numUnits, int lastCellUnit, int lastInputUnit, int biasUnit,
                         int numBlocks, int[] blockSize, int numPositionBins, double initRange) {
        
        final int randfactor = 1000000;
        
        this.numUnits          = numUnits;
        this.lastCellUnit      = lastCellUnit;
        this.lastInputUnit     = lastInputUnit;
        this.biasUnit          = biasUnit;
        this.numBlocks         = numBlocks;
        this.blockSize         = blockSize;
        this.numPositionBins   = numPositionBins;
        
        weightMatrix = new double[numUnits][numUnits];
    
        wmdimx       = numUnits;
        wmdimyoffset = biasUnit + 1;
        wmdimy       = numUnits - wmdimyoffset;
        
        System.err.println();
        System.err.print("Weight matrix initialisation.. ");
        
        
        // Init (all weights are set to zero here)
        for (int i = 0; i < numUnits; i++) {
            for (int j = 0;j <= lastInputUnit; j++) {
                
                              	
                weightMatrix[i][j] = seprand(randfactor * 2) - randfactor;
                weightMatrix[i][j] /= randfactor;
                weightMatrix[i][j]*=initRange;
            	
                
            	//weightMatrix[i][j] = 0;

            }
            for (int j = biasUnit;j < numUnits; j++) {
                
            	/*
                weightMatrix[i][j] = seprand(randfactor * 2) - randfactor;
                weightMatrix[i][j] /= randfactor;
                weightMatrix[i][j]*=initRange;
                */
            	
                weightMatrix[i][j] = 0;
                
            }
        }
        
        System.err.println("okay");
        
        
    }
    
    /**
     * Reads a weight matrix from file system
     * 
     */
    public void readWeightMatrix() {
    	
    	/* reset weight matrix to zero */
    	for (int i = 0; i < this.wmdimx; i++) {
    		for (int j = 0; j < this.wmdimx; j++) {
    			weightMatrix[i][j] = 0;
    		}
    	}
    	
    	Scanner s = null;
    	
    	try {
        	
            s = new Scanner(new BufferedReader( new FileReader( weightFile ) ));

            String dim = s.next("dim");
            
            if (!"dim".equals(dim)) {
            	
            	System.err.println("Error in weight matrix. Wrong dimension header");
            	System.exit(100);
            	
            }
            
            this.wmdimy = s.nextInt();
            this.wmdimx = s.nextInt();
            this.wmdimyoffset = this.wmdimx - this.wmdimy;

            for ( int i = 0; i < wmdimy; i++ ) {
          	  
                int j;
                
                int k = wmdimyoffset + i;
                
                for (j = 0; j < wmdimx; j++) {
              	  
                    weightMatrix[k][j] = s.nextDouble();
                
                }
                
            }
  
            s.close();
            
          }
          
          catch ( FileNotFoundException fnfe ) {
        	  System.err.println( "Weight matrix file not found" );
        	  System.exit(101);
          }
          
          catch (InputMismatchException ime ) {
        	  
        	  System.err.println( "Error in weight matrix, input mismatch" );
        	  System.exit(102);
        	  
          }
          
          finally {
            
        	  s.close();
        	  
          }
        
    }
    
    /**
     * writes the current weight matrix to a file
     * 
     */
    public void writeWeightMatrix() {
        
        BufferedWriter out = null;

        try {
        	
          out = new BufferedWriter( new FileWriter( weightFile ) );
          
          out.write("dim " + String.valueOf(wmdimy) + " " + String.valueOf(wmdimx) + "\n");

          for ( int i = 0; i < wmdimy; i++ ) {
        	  
              int j;
              
              int k = wmdimyoffset + i;
              
              for (j = 0; j < wmdimx - 1; j++) {
            	  
                  out.write(new DecimalFormat("#.####").format(weightMatrix[k][j]) + " ");
              
              }
              
              out.write(new DecimalFormat("#.####").format(weightMatrix[k][j]));
              
              out.write("\n");
              
          }
          
          out.write("\n");
          
          out.close();
          
        }
        
        catch ( IOException e ) {
          System.err.println( e );
        }
        
        finally {
          try {
            if ( out != null ) out.close();
          } catch ( IOException e ) { }
        }
        
    }
    
    
    
    /**
     * write a better readable weight matrix
     * 
     */
    public void writeReadableWeightMatrix() {
        
        // TODO: Better formatting
        
        BufferedWriter out = null;
        
        String rweightfile = "readable_" + weightFile;

        try {
          
          out = new BufferedWriter( new FileWriter( rweightfile ) );
          
          out.write("dim " + String.valueOf(wmdimy) + ":" + String.valueOf(wmdimx));
          out.newLine();

          for ( int i = 0; i < wmdimy; i++ ) {
              int j;
              int k = wmdimyoffset + i;
              for (j = 0; j < wmdimx - 1; j++) {
                  out.write("(" + String.valueOf(k) + "," + String.valueOf(j) + "): ");
                  out.write(new DecimalFormat("#.####").format(weightMatrix[k][j]) + " ");
              }
              out.write("(" + String.valueOf(k) + "," + String.valueOf(j) + "): ");
              out.write(new DecimalFormat("#.####").format(weightMatrix[k][j]));
              out.newLine();
          }
          out.newLine();
          out.close();
        }
        catch ( IOException e ) {
          System.err.println( e );
        }
        finally {
          try {
            if ( out != null ) out.close();
          } catch ( IOException e ) { }
        }
        
    }
    
 
    /**
     * Computes weight updates
     * 
     * A regularization is done on the input weights with the Gonnet500 matrix
     * 
     * @param DW
     */
    public void weightUpdate(double[][] DW, ArrayList<Integer> usedInputIndices) {
        
        //final double wdecay = 0.00001;
        
        /* output units */

        /* weights fixed to +- 1.0 */
    	
        /* memory cells with gates */
        
        int i = biasUnit + 1;
        
        for (int u = 0; u < numBlocks; u++) {

        	/* input gate */
        	
        	/* input gate is fixed to 1.0 */
           
            i++;

            /* output gate */
            
            /* og fixed to 1.0 */
            
            /* memory cell */
            for (int v = 0; v < blockSize[u]; v++) {
                
                i++;          
                
                /* only input weights from k-mers found in sequence
                 * need an update */
                for (int j = 0; j < usedInputIndices.size(); j++) {
                	
                	int uii = usedInputIndices.get(j);
                                                    
                    //weightMatrix[i][uii] +=  DW[i][uii] + wdecay * weightMatrix[i][uii];
                    weightMatrix[i][uii] +=  DW[i][uii];
                        
                }
                
                for (int j = 0; j < numPositionBins; j++) {
                	
                	int idx = biasUnit - numPositionBins + j;
                	//weightMatrix[i][idx] +=  DW[i][idx] + wdecay * weightMatrix[i][idx];
                	weightMatrix[i][idx] +=  DW[i][idx];
                	
                }
                
                
                
            }
            
            i++;
            
        }
        
    }

    
    
    /**
     * set the weight file name
     * 
     * @param weightFile
     */
    public void setWeightfile(String weightFile) {
        this.weightFile = weightFile;
    }

    /**
     * get the weight matrix
     * 
     * @return weightMatrix
     */
    public double[][] getWeightMatrix() {
        return weightMatrix;
    }

    /**
     * Computes a random integer
     * 
     * @param k
     * @return
     */
    int seprand(int k) {
        long l;
        int f;
        Random rand = new Random();
        l = rand.nextLong();
        f = (int)l % k;
        return(f);
      }


}
